import StringIO
from django.template import loader
from django.template.response import TemplateResponse
from django.utils.encoding import smart_str

from ereports.templatetags.ereports import with_widget


class ReportRender(object):
    def __init__(self, report):
        super(ReportRender, self).__init__()
        self.report = report

    def render(self, request, context, **response_kwargs):
        context['total_columns_num'] = len(self.report.headers)
        return super(ReportRender, self).render(request, context, **response_kwargs)


class TemplateRender(object):
    """
    A mixin that can be used to render a template.
    """
    template_name = None
    response_class = TemplateResponse
    content_type = None

    def render(self, request, context, **response_kwargs):
        """
        Returns a response, using the `response_class` for this
        view, with a template rendered with the given context.

        If any keyword arguments are provided, they will be
        passed to the constructor of the response class.
        """
        template = loader.get_template(self.active_template_name)
        # context = self.resolve_context(self.context_data)
        content = template.render(context)
        return content


class XlsRender(object):
    """
    A mixin that can be used to render an xls file.
    """
    content_type = 'application/vnd.ms-excel'

    def _get_styles(self, context):
        from xlwt import easyxf

        FORMATS = {
            'DateColumn': 'DD-MMM-YYYY',
            'DateTimeColumn': 'DD MMD YY hh:mm',
            'TimeColumn': 'hh:mm',
            'IntegerColumn': '#,##',
            'DecimalColumn': '#,##0.00',
            'BooleanColumn': 'general',
            'CurrencyColumn': '"$"#,##0.00);[Red]("$"#,##0.00)',
        }

        styles = {}
        for col in context['report'].display_order():
            col_style = context['report'].get_column_by_name(col)
            styles[col] = easyxf(num_format_str=FORMATS.get(col_style.__class__.__name__, 'general'))

        return styles

    #TODO:
    # handle columns Email, Formulas, etc
    # n = "HYPERLINK"
    # ws.write_merge(1, 1, 1, 10, Formula(n + '("http://www.irs.gov/pub/irs-pdf/f1000.pdf";"f1000.pdf")'), h_style)
    # ws.write(1, 0, xlwt.Formula('A1*B1')) # Should output "10" (A1[5] * A2[2])
    # ws.write(1, 1, xlwt.Formula('SUM(A1,B1)')) # Should output "7" (A1[5] + A2[2])

    def render(self, request, context, **response_kwargs):
        from xlwt import Workbook, XFStyle, easyxf

        w = Workbook(encoding='utf-8')

        ws = w.add_sheet('Report')
        style = XFStyle()

        row = 0
        heading_xf = easyxf('font:height 200; font: bold on; align: wrap on, vert centre, horiz center')
        ws.write(row, 0, '#', style)

        for col, fieldname in enumerate(context['report'].headers, start=1):
            ws.write(row, col, str(fieldname), heading_xf)
            ws.col(col).width = 5000
        ws.row(row).height = 500

        # we have to prepare all the styles before going into the loop
        # to avoid the "More than 4094 XFs (styles)" Error
        styles = self._get_styles(context)
        for rownum, data in enumerate(context['report']):
            ws.write(rownum + 1, 0, rownum + 1)
            for idx, (fieldname, rowvalue) in enumerate(data.items()):
                style = styles[rowvalue.column.name]
                try:
                    ws.write(rownum + 1, idx + 1, with_widget(rowvalue, format='xls'), style)
                except Exception:
                    #logger.warning("TODO refine this exception: %s" % e)
                    ws.write(rownum + 1, idx + 1, smart_str(with_widget(rowvalue)), style)

        f = StringIO.StringIO()
        w.save(f)
        f.seek(0)

        return f.read()



# -*- coding: utf-8 -*-
import datetime
from decimal import Decimal, InvalidOperation
from django import template
from django.contrib.humanize.templatetags.humanize import intcomma
from django.core.urlresolvers import reverse
from django.db import models
from django.db.models import BooleanField
from django.db.models.fields.related import ForeignKey
from django.template.base import Template
from django.template.defaultfilters import floatformat, date
import pprint
from django.utils.encoding import smart_str
from django.utils.safestring import mark_safe
from kiowa.utils import get_attr as recursive_getattr


register = template.Library()


def _get_field(Model, field_path):
    """
    get a Model and a path to a attribute, return the field


    >>> a = _get_field(Payslip, 'contract.employee.health_ins_type')
    >>> print a
    Health Insurance Type
    """
    parts = field_path.split('.')
    target = parts[0]
    if target in Model._meta.get_all_field_names():
        field_object, model, direct, m2m = Model._meta.get_field_by_name(target)
        if isinstance(field_object, models.fields.related.ForeignKey):
            if parts[1:]:
                return _get_field(field_object.rel.to, '.'.join(parts[1:]))
            else:
                return field_object
        else:
            return field_object
    return None


@register.simple_tag
def header(model, field_name):
    field = _get_field(model, field_name)
    if field and hasattr(field, 'verbose_name'):
        label = field.verbose_name
    elif '|' in field_name:
        name, __ = field_name.split('|')
        label = name.replace('.', ' ').replace('_', ' ').title()
    else:
        label = field_name.replace('.', ' ').replace('_', ' ').title()
    return smart_str(label)


@register.simple_tag(takes_context=True, name='cell')
def process_queryset_field(context, obj, field_name, raw=False):
    context['current_row'] = obj
    if raw:
        if '|' in field_name:
            field_name, __ = field_name.split('|')
        if '#' in field_name:
            field_name, __ = field_name.split('|')

        attr = recursive_getattr(obj, field_name)
        Field = _get_field(obj.__class__, field_name)
        if callable(attr):
            return attr()
        elif issubclass(Field.__class__, BooleanField):
            return {True:'yes', False:'no'}[attr]
        return attr
    else:
        if field_name.startswith('{'):
            pattern = "{%% load ereports %%}%s" % field_name
        elif '|' in field_name:
            pattern = "{%% load ereports %%}{{current_row.%s}}" % field_name
        else:
            pattern = "{%% load ereports %%}{{current_row.%s}}" % field_name

        t = Template(pattern)
        value = t.render(context)
        return value or '&nbsp;'

@register.filter()
def currency(dollars):
    dollars = round(float(dollars), 2)
    return "%s%s" % (intcomma(int(dollars)), ("%0.2f" % dollars)[-3:])


def isNumeric(s):
    try:
        return int(s) or True
    except (ValueError, TypeError):
        pass
    try:
        Decimal(s) or True
    except (ValueError, TypeError, InvalidOperation):
        pass
    try:
        return float(s) or True
    except (ValueError, TypeError):
        pass
    return False


@register.simple_tag
def formatvalue(value):
    if isinstance(value, (Decimal, float)):
        ret = "<div class='currency'>%s</div>" % currency(value)
    elif isinstance(value, (int,)):
        ret = "<div class='int'>%s</div>" % value
    elif isinstance(value, (datetime.datetime, datetime.date)):
        ret = "<div class='date nobr'>%s</div>" % date(value, 'd b Y')
    elif isinstance(value, basestring) and isNumeric(value):
        ret = "<div class='numeric'>%s</div>" % value
    elif value is None or unicode(value).strip() == '':
        ret = '&nbsp;'
    else:
        ret = value
    return mark_safe(ret)

@register.filter(name='formatvalue')
def formatvalue_filter(value):
    return formatvalue(value)

@register.filter(name='formatvalue')
def formatvalue_filter(value):
    return mark_safe(formatvalue(value))


@register.filter('str')
def _str(target):
    return str(target)


@register.filter('int')
def _int(target):
    return int(target)


@register.filter('float')
def _float(target):
    return float(target)


@register.simple_tag
def field_link(owner, field):
    if owner:
        label = '%s.%s' % (owner, field.name)
    else:
        label = field.name

    args = {'class': 'field',
            'label': label,
            'title': '',
            'owner': None}

    if isinstance(field, ForeignKey):
        opts = field.rel.to._meta
        args.update({'class': 'inspect closed field',
                     'title': reverse('get_model_meta', args=[label, opts.app_label, opts.object_name.lower()])})
    return '<div class="%(class)s" title="%(title)s">%(label)s</div>' % args


@register.simple_tag(takes_context=True)
def make_total(context, value, var_name):
    var = var_name[1:]
    if var_name[0] == '+':
        partial = context.get(var, 0)
        partial = partial + value
        context[var] = partial
        return floatformat(value, 2)
    elif var_name[0] == '=':
        return floatformat(context[var], 2)


@register.simple_tag(takes_context=True)
def qs(context, varname, default=None):
    filter_target = context['request'].GET.get(varname, default)
    return filter_target


@register.simple_tag(takes_context=True)
def edebug(context):
    ret = {}
    for i in context:
        for k, v in i.items():
            if k not in ['request', 'LANGUAGES', 'sql_queries', 'headers',
                         'STATIC_URL', 'LANGUAGE_BIDI', 'LANGUAGE_CODE',
                         'csrf_token', 'session_cookie_alert',
                         'session_cookie_age_fmt',
                         'session_cookie_expire_fmt',
                         'session_cookie_age',
                         'uitheme', 'kiowamedia', 'browser', ]:
                ret[k] = v

                #    ret.append('%s: %s' % (k, str(context[k])))

    return pprint.pformat(ret)


@register.filter
def running_total(qs, fieldname):
    values = [recursive_getattr(d, fieldname) for d in qs]

    def is_summable(value):
        try:
            float(value)
            return True
        except (ValueError, TypeError):
            return False

    return sum(map(float, filter(is_summable, values)))


@register.filter
def payroll_total(qs, fieldname):
    values = [recursive_getattr(d, fieldname) for d in qs]
    try:
        return formatvalue(sum(map(float, values)))
    except (ValueError, TypeError):
        return ''


@register.filter
def currencyformat(value, thousand_sep=',', decimal_sep='.'):
    """
    Returns the currency format of the value param using specified separators

    Usage: {% value_to_format|currencyformat %}
    """
    integer_part_string = str(int(value))
    f = lambda x, y, list=[]: f(x[:-y], y, [(x[-y:])] + list) if x else list
    integer_part = thousand_sep.join(f(integer_part_string, 3))
    return "%s%s%s" % (integer_part, decimal_sep, ("%0.2f" % value)[-2:])


@register.simple_tag
def summarize(group, fieldname):
    return formatvalue(group.sum(fieldname))


@register.simple_tag(name='get_attr')
def get_attr_tag(target, attrname):
    return recursive_getattr(target, attrname)


@register.simple_tag
def get_formatted_value(target, attrname):
    return formatvalue(recursive_getattr(target, attrname))


@register.simple_tag
def astoday(object, method_name):
    m = recursive_getattr(object, method_name)
    return m(datetime.date.today())


@register.filter
def div_mod(dividend, divisor):
    quotient, remainder = divmod(dividend, divisor)
    return quotient, remainder


@register.filter
def yes_no(value):
    return {True: 'Yes', False: 'No', None: 'N/A'}.get(value, 'N/A')




# -*- coding: utf-8 -*-
import logging
from collections import namedtuple
from ereports.engine.cache import monitor_model
from ereports.utils import fqn, get_attr

logger = logging.getLogger(__name__)

_Entry = namedtuple('Entry', "classname,type")


class Registry(object):
    """
        Registry of all known Report(s)
    """

    def __init__(self):
        super(Registry, self).__init__()
        self.reports = []
        self.monitors = []

    def register(self, report_class):
        model = get_attr(report_class, 'datasource.model', None)
        if model is None:
            logger.error("Unable to monitor {0}".format(report_class))

        elif model not in self.monitors:
            self.monitors.append(model)
            monitor_model(model)

        self.reports.append(_Entry(fqn(report_class), report_class))

    def __iter__(self):
        for el in self.reports:
            yield el

    def __getitem__(self, item):
        if isinstance(item, int):
            return self.reports[item]
        elif isinstance(item, basestring):
            return filter(lambda el: el.classname == item, self)[0][1]

    def __contains__(self, item):
        return item in [r.classname for r in self]

    def choices(self):
        return [(el.classname, el.classname) for el in self]

    def get(self, class_name, default=None):
        if class_name in self:
            return self[class_name]
        return default


registry = Registry()



import datetime
import pytest

from django_webtest import WebTest
from django import forms
from django.core.exceptions import ValidationError
from django.utils.datastructures import MultiValueDict

from ereports.engine.config import reportform_factory, DateRange, DateRangeField, DateRangeInput, \
    ColumnsChoiceField, base_has_attr, get_attr_from_bases, get_field_from_bases, FilterSpec, UIDateInput, \
    MultiChoiceWidget
from ereports.tests.app.reports import SimpleDemoReport, SimpleDateReport, FilterSpecDemoReport, SimpleDemoExtraReport


class FakeBase(forms.Form):
    name = 'Fake'
    field_name = forms.CharField(max_length=255)


def test_base_has_attr():
    assert base_has_attr((FakeBase, ), 'name')
    assert base_has_attr((object(), FakeBase, ), 'name')
    assert not base_has_attr((FakeBase, ), 'wrong')
    assert not base_has_attr((object(), FakeBase, ), 'wrong')


def test_get_attr_from_bases():
    assert get_attr_from_bases((FakeBase,), 'name') == 'Fake'
    assert get_attr_from_bases((FakeBase,), 'wrong', 'default') == 'default'

    with pytest.raises(UnboundLocalError):
        get_attr_from_bases((FakeBase,), 'wrong')


def test_get_field_from_bases():
    assert isinstance(get_field_from_bases((FakeBase,), 'field_name'), forms.CharField)
    assert get_field_from_bases((FakeBase,), 'wrong', 'default') == 'default'

    with pytest.raises(UnboundLocalError):
        get_field_from_bases((FakeBase,), 'wrong')


class TestFilterSpec(WebTest):
    def test_init(self):
        f = FilterSpec()
        self.assertEqual(f._value, {})

        f = FilterSpec(value='abc')
        self.assertEqual(f._value, 'abc')

    def test_get_filter_description(self):
        f = FilterSpec(value='abc')
        self.assertEqual(f.get_filter_description('char'), "abc")

    def test_get_filter(self):
        f = FilterSpec(value='abc')
        self.assertEqual(f.get_filter(), 'abc')

    def test_nonzero(self):
        f = FilterSpec(value='abc')
        self.assertEqual(bool(f), True)

        f = FilterSpec()
        self.assertEqual(bool(f), False)

        f = FilterSpec(value='')
        self.assertEqual(bool(f), False)


class TestConfigurationForm(WebTest):
    def test_init(self):
        r = SimpleDemoReport.as_report()
        c = reportform_factory(r)
        self.assertIn('_format', c.base_fields)

    def test_get_lookup_fields(self):
        r = SimpleDemoReport.as_report()
        form = reportform_factory(r, (r.config_form_class,), filtering=['integer1'])()
        f = form.get_lookup_fields()
        self.assertEqual(['integer1'], f)

    def test_get_filters(self):
        r = SimpleDemoReport.as_report()
        form = reportform_factory(r, (r.config_form_class,), filtering=['integer1'])()
        form.cleaned_data = {
            'integer1': 1,
            'char': ['a', 'b', 'c']
        }
        f = form.get_filters()
        self.assertSequenceEqual(([], {'integer1': 1}), f)

        form.cleaned_data = {
            'integer1': [1, 2, 3]
        }
        f = form.get_filters()
        self.assertSequenceEqual(([], {'integer1__in': [1, 2, 3]}), f)

    def test_get_filters_filterspec(self):
        r = FilterSpecDemoReport.as_report()
        form = reportform_factory(r, (r.config_form_class,))()
        form.cleaned_data = {
            'date_range': FilterSpec('2001')
        }
        f = form.get_filters()
        self.assertSequenceEqual((['2001'], {}), f)

    def test_get_filters_filterspec_dict(self):
        r = FilterSpecDemoReport.as_report()
        form = reportform_factory(r, (r.config_form_class,))()
        form.cleaned_data = {
            'date_range': FilterSpec({1: '2001'})
        }
        f = form.get_filters()
        self.assertSequenceEqual(([], {1: '2001'}), f)

    def test_get_filters_date_range(self):
        r = SimpleDateReport.as_report()
        form = reportform_factory(r, (r.config_form_class,))()
        form.cleaned_data = {
            'date_range': DateRange(['2000-01-01', '2001-01-01']),
            'date': '2000-01-01'
        }
        f = form.get_filters()
        self.assertEqual({'date_range__gt': '2000-01-01',
                          'date_range__lt': '2001-01-01',
                          'date': '2000-01-01'}, f[1])

    def test_get_filters_summary(self):
        r = SimpleDemoReport.as_report()
        form = reportform_factory(r, (r.config_form_class,))()
        form.cleaned_data = {
            'integer1': 1,
        }
        self.assertEqual(form.get_filters_summary(), {u'Integer #1': 1})

        form.cleaned_data = {
            'integer1': [1, 2, 3]
        }
        form.get_filters()
        self.assertEqual(form.get_filters_summary(), {u'Integer #1': [1, 2, 3]})

    def test_get_filters_summary_date_range(self):
        r = SimpleDateReport.as_report()
        form = reportform_factory(r, (r.config_form_class,))()
        form.cleaned_data = {
            'date_range': DateRange(['2000-01-01', '2001-01-01']),
            'date': '2000-01-01'
        }
        self.assertEqual(form.get_filters_summary(),
                         {u'Date': '2000-01-01', u'Date Range': ['2000-01-01', '2001-01-01']})

    def test_get_filters_summary_filterspec(self):
        r = FilterSpecDemoReport.as_report()
        form = reportform_factory(r, (r.config_form_class,))()
        form.cleaned_data = {
            'date_range': FilterSpec('2001')
        }
        self.assertEqual(form.get_filters_summary(), {u'Date Range': '2001'})

    def test_get_report_attributes(self):
        r = SimpleDateReport.as_report()
        form = reportform_factory(r, (r.config_form_class,))()

        self.assertIn('_report_list_display', form.base_fields)
        self.assertIsInstance(form.base_fields['_report_list_display'], ColumnsChoiceField)

        form.cleaned_data = {
            '_report_list_display': 'abc',
            '_report_group_by': 'random',
        }
        a = form.get_report_attributes()
        self.assertEqual({'list_display': 'abc', 'group_by': 'random'}, a)

    def test_get_report_attributes_extras(self):
        r = SimpleDemoExtraReport.as_report()
        form = reportform_factory(r, (r.config_form_class,))()

        self.assertIn('_extras_data', form.base_fields)

        form.cleaned_data = {
            '_report_list_display': 'char',
            '_extras_data': 'more'
        }
        a = form.get_report_attributes()
        self.assertEqual({'list_display': 'char', 'extras': {'data': 'more'}}, a)

    def test_clean__report_order_by(self):
        r = SimpleDateReport.as_report()
        form = reportform_factory(r, (r.config_form_class,))()
        form.cleaned_data = {
            '_report_order_by': 'user.first_name',
        }
        o = form.clean__report_order_by()
        self.assertEqual(['user__first_name'], o)

        form.cleaned_data = {
            '_report_order_by': 'user.first_name, user.last_name',
        }
        o = form.clean__report_order_by()
        self.assertEqual(['user__first_name', 'user__last_name'], o)

    def test_html_output(self):
        r = SimpleDateReport.as_report()
        form = reportform_factory(r, (r.config_form_class,))()
        self.assertEqual(form.fields.keys(),
                         ['date_range', '_report_group_by', 'date', '_format', '_report_list_display'])
        form.as_p()
        self.assertEqual(form.fields.keys(), ['date', '_report_group_by', '_format', '_report_list_display'])

    def test_report_order_by(self):
        r = SimpleDemoReport.as_report()
        form = reportform_factory(r, (r.config_form_class,), order_by=[('date', 'Date'), ])()
        self.assertIn('_report_order_by', form.base_fields)
        self.assertIsInstance(form.base_fields['_report_order_by'], forms.ChoiceField)

    def test_report_groupby(self):
        r = SimpleDemoReport.as_report()
        form = reportform_factory(r, (r.config_form_class,), groupby=[('date', 'Date'), ])()
        self.assertIn('_report_group_by', form.base_fields)
        self.assertIsInstance(form.base_fields['_report_group_by'], forms.ChoiceField)

    def test_report_groupby_from_base(self):
        r = SimpleDateReport.as_report()
        form = reportform_factory(r, (r.config_form_class,))()
        self.assertIn('_report_group_by', form.base_fields)
        self.assertIsInstance(form.base_fields['_report_group_by'], forms.ChoiceField)

    def test_clean__report_group_by(self):
        r = SimpleDemoReport.as_report()
        form = reportform_factory(r, (r.config_form_class,), groupby=[('date', 'Date'), ])()
        form.cleaned_data = {
            'date': '2001-01-01',
            '_report_group_by': 'date'
        }
        self.assertIsNone(form.clean__report_group_by())

        form.cleaned_data = {
            'date': '2001-01-01',
            '_report_group_by': 'date,internal'
        }
        group, internal_order = form.clean__report_group_by()
        self.assertEqual(group, 'date')
        self.assertEqual(internal_order, 'internal')


class TestDateRangeField(WebTest):
    def test_to_python(self):
        d = DateRangeField()

        r = d.to_python(['2000-01-01', '2001-01-01'])
        self.assertIsInstance(r, DateRange)
        self.assertEqual(r, DateRange([datetime.date(2000, 1, 1), datetime.date(2001, 1, 1)]))

        with self.assertRaises(ValidationError):
            r = d.to_python(['2000', '2001-01-01'])

        with self.assertRaises(ValidationError):
            r = d.to_python(['2000-01-01', '2001'])

    def test_clean(self):
        d = DateRangeField()
        r = d.clean(['2000-01-01', '2001-01-01'])

        self.assertIsInstance(r, DateRange)
        self.assertEqual(r, DateRange([datetime.date(2000, 1, 1), datetime.date(2001, 1, 1)]))

        with self.assertRaises(ValidationError):
            r = d.clean(['2000', '2001-01-01'])

        with self.assertRaises(ValidationError):
            r = d.clean(['2000-01-01', '2001'])


class TestDateRangeInput(WebTest):
    def test_value_from_datadict(self):
        d = DateRangeInput()
        data = {
            'date_start': '2000-01-01',
            'date_end': '2001-01-01',
        }
        name = 'date'

        v = d.value_from_datadict(data, None, name)
        self.assertEqual(v, ['2000-01-01', '2001-01-01'])

    def test_render(self):
        d = DateRangeInput()
        v = ['2000-01-01', '2001-01-01']
        r = d.render('date', v, {'id': 'id_date'})
        self.assertIn('name="date_start"', r)
        self.assertIn('name="date_end"', r)

        v = [None, '2001-01-01']
        r = d.render('date', v, {'id': 'id_date'})
        self.assertIn('name="date_start"', r)
        self.assertIn('name="date_end"', r)
        self.assertNotIn('value="2000-01-01"', r)
        self.assertIn('value="2001-01-01"', r)

        v = ['2000-01-01', None]
        r = d.render('date', v, {'id': 'id_date'})
        self.assertIn('name="date_start"', r)
        self.assertIn('name="date_end"', r)
        self.assertIn('value="2000-01-01"', r)
        self.assertNotIn('value="2001-01-01"', r)

        v = '2000-01-01'
        r = d.render('date', v, {'id': 'id_date'})
        self.assertIn('name="date_start"', r)
        self.assertIn('name="date_end"', r)
        self.assertNotIn('value="2000-01-01"', r)


class TestUIDateInput(WebTest):
    def test_render(self):
        d = UIDateInput()
        v = '2001-01-01'
        name = 'date'
        r = d.render(name, v, {})
        self.assertIn('name="date"', r)
        self.assertIn('class="uidate2"', r)


class TestMultiChoiceWidget(WebTest):
    def test_value_from_datadict(self):
        m = MultiChoiceWidget()
        d = MultiValueDict({'name': [1, 2, 3]})
        self.assertEqual(m.value_from_datadict(d, {}, 'name'), [1, 2, 3])



SECRET_KEY = '\x01\xff[\xa6\x95\x9daG&\xa8%8\x00"~\xb7W\xbfm\x01\x8269\xa1'
TIME_ZONE = 'Asia/Bangkok'
DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': ':memory:'
    },
    'interface': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': ':memory:'
    }
}
STATIC_URL = '/static/'

INSTALLED_APPS = (
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.sites',
    'django.contrib.messages',
    'django.contrib.staticfiles',
    'ereports',
    'ereports.tests.app',
)

LOGGING = {
    'version': 1,
    'disable_existing_loggers': False,
    'formatters': {
        'full': {
            'format': '%(levelname)-8s: %(asctime)s %(module)s %(process)d %(thread)d %(message)s'
        },
        'verbose': {
            'format': '%(levelname)-8s: %(asctime)s %(name)-25s %(message)s'
        },
        'simple': {
            'format': '%(levelname)-8s %(asctime)s %(name)-25s %(funcName)s %(message)s'
        },
        'debug': {
            'format': '%(levelno)s:%(levelname)-8s %(name)s %(funcName)s:%(lineno)s:: %(message)s'
        }
    },
    'filters': {
        'require_debug_false': {
            '()': 'django.utils.log.RequireDebugFalse'
        }
    },
    'handlers': {
        'null': {
            'level': 'DEBUG',
            'class': 'django.utils.log.NullHandler'
        },
        'console': {
            'level': 'DEBUG',
            'class': 'logging.StreamHandler',
            'formatter': 'debug'
        }
    },
    'loggers': {
        'concurrency': {
            'handlers': ['null'],
            'propagate': False,
            'level': 'DEBUG'
        }
    }
}



# -*- coding: utf-8 -*-
from importlib import import_module
from optparse import make_option
import sys
from django.contrib.contenttypes.models import ContentType
from django.core.management.base import BaseCommand
from django.db.utils import IntegrityError
from ereports.models import ReportConfiguration, ReportGroup, keygen
from ereports.utils import import_by_path


class Command(BaseCommand):
    args = ''
    help = 'register Report classes and create one ReportConfiguration per each'
    option_list = BaseCommand.option_list + (
        make_option('-m', '--module',
                    action='store',
                    dest='module',
                    default=None,
                    help='extra module to scan'),
        make_option('-o', '--overwrite',
                    action='store_true',
                    dest='overwrite',
                    default=False,
                    help='overwrite existing imported reports'),
        make_option('-z', '--clean',
                    action='store_true',
                    dest='clean',
                    default=False,
                    help='remove existing reports in the selected group'),
        make_option('-g', '--group',
                    action='store',
                    dest='group',
                    default='Standard Report',
                    help='group name the reports will belong to'),
    )

    def handle(self, *args, **options):
        stdout = options.get('stdout', sys.stdout)
        module = options.get('module')
        overwrite = options.get('overwrite')
        clean = options.get('clean')
        groupname = options.get('group')
        verbosity = int(options.get('verbosity'))

        group, __ = ReportGroup.objects.get_or_create(name=groupname)

        from ereports.engine.registry import registry
        from ereports.engine.report import BaseReport

        if module:
            import_module(module.strip())
        if clean:
            ReportConfiguration.objects.filter(group=group).delete()

        # entry is (fqn(report_class), report_class)
        for entry in registry:
            report = import_by_path(entry.classname)

            if not report is BaseReport:
                report = report.as_report()
                if verbosity >= 1:
                    stdout.write("Importing...%s" % entry.classname)
                model = report.model or report.datasource.model
                name = report.title
                i = 0
                while True:
                    try:
                        cfg, isnew = ReportConfiguration.objects.get_or_create(group=group,
                                                                               report_class=entry.classname,
                                                                               defaults={'name': name})
                        break
                    except IntegrityError:
                        i += 1
                        name = "{0.title}_{1}".format(report, i)

                if isnew or overwrite:
                    cfg.columns = "\r\n".join([col.name for col in report.datasource.columns])
                    cfg.target_model = ContentType.objects.get_for_model(model)
                    cfg.published = True
                    cfg.filtering = "\r\n".join(report.list_filter)

                    cfg.cache_key = keygen()
                cfg.save()
                if verbosity >= 1:
                    stdout.write("...Done\n")



# -*- coding: utf-8 -*-
from django import forms
from django.contrib.auth.models import User
from django.core.exceptions import ImproperlyConfigured
from django.http import QueryDict
from django.test.testcases import TestCase

from ereports.filtering import QueryField, FilterQuerysetMixin, ANY_LOOKUP_ALLOWED
from ereports.tests.fixtures import get_fake_request
from ereports.utils import StartWithList


class FakeSearchForm(forms.Form):
    arg1 = forms.CharField(max_length=150)
    arg2 = forms.CharField(max_length=150)
    _random = forms.CharField(max_length=20)
    username = QueryField(max_length=20)


class TestQueryField(TestCase):
    def test_init(self):
        q = QueryField()
        self.assertIsInstance(q, QueryField)


class TestFilterQuerysetMixin(TestCase):
    def setUp(self):
        self.mixin = FilterQuerysetMixin()
        self.mixin.model = User
        self.request = get_fake_request()

    def test_get_params(self):
        params = self.mixin.get_params(self.request)
        self.assertEqual(params, {u'arg1': u'test', u'arg2': u'two'})

    def test_get_querydict(self):
        querydict = self.mixin.get_querydict(self.request)
        self.assertIsInstance(querydict, QueryDict)

    def test_get_search_form(self):
        form = self.mixin.get_search_form()
        self.assertIsInstance(form, self.mixin.search_form_class)

    def test_get_query_help_message(self):
        self.assertEqual(self.mixin.get_query_help_message(), "")

        self.mixin.search_fields = ['nonexistant']
        with self.assertRaises(ImproperlyConfigured):
            self.mixin.get_query_help_message()

        self.mixin.search_fields = ['username']
        self.assertEqual(self.mixin.get_query_help_message(), 'Username')

        self.mixin.search_fields = ['username', 'password']
        self.assertEqual(self.mixin.get_query_help_message(), 'Username,Password')

    def test_query_param(self):
        self.assertEqual(self.mixin.query_param, None)

        self.mixin.search_form_class = FakeSearchForm
        self.assertEqual(self.mixin.query_param, 'username')

    def test_query(self):
        self.mixin.params = self.mixin.get_params(self.request)
        self.assertEqual(self.mixin.query, '')

        self.mixin.method = 'POST'
        self.mixin.params = self.mixin.get_params(self.request)
        self.mixin.search_form_class = FakeSearchForm
        self.assertEqual(self.mixin.query, u'random')

    def test_get_lookup_fields(self):
        lk = self.mixin.get_lookup_fields()
        self.assertEqual(lk, [])
        self.assertIsInstance(lk, StartWithList)

        self.mixin.search_form_class = FakeSearchForm
        lk = self.mixin.get_lookup_fields()
        self.assertEqual(lk, ["arg1", "arg2"])

        self.mixin.lookup_fields = ANY_LOOKUP_ALLOWED
        self.assertEqual(self.mixin.get_lookup_fields(), ANY_LOOKUP_ALLOWED)

    def test_get_filters(self):
        self.assertEqual(self.mixin._get_filters(), {})

        self.mixin.search_form_class = FakeSearchForm
        self.mixin.params = self.mixin.get_params(self.request)
        self.mixin.querydict = self.mixin.get_querydict(self.request)
        self.assertEqual(self.mixin._get_filters(), {u"arg1": u"test", "arg2__in": ["one", "two"]})

    def test_filter_qs(self):
        qs = User.objects.all()
        res = self.mixin._filter_qs(qs)
        self.assertEqual(len(qs), len(res))

        self.mixin.method = 'POST'
        self.mixin.search_form_class = FakeSearchForm
        self.mixin.params = self.mixin.get_params(self.request)
        self.mixin.querydict = self.mixin.get_querydict(self.request)

        self.mixin.search_fields = ['username']
        self.mixin.params = {u'username': 'Sax'}
        res = self.mixin._filter_qs(qs)
        self.assertEqual(len(res), 1)

        self.mixin.search_fields = ['^username']
        self.mixin.params = {u'username': 'sa'}
        res = self.mixin._filter_qs(qs)
        self.assertEqual(len(res), 1)

        self.mixin.search_fields = ['=username']
        self.mixin.params = {u'username': 'sax'}
        res = self.mixin._filter_qs(qs)
        self.assertEqual(len(res), 1)

        #NOTE Full-text search not implemented for testing db backend
        # self.mixin.search_fields = ['@username']
        # self.mixin.params = {u'username': 'sax'}
        # res = self.mixin._filter_qs(qs)
        # self.assertEqual(len(res), 1)



from __future__ import absolute_import
import datetime
import pytest
from decimal import Decimal
from django.contrib.auth.models import User
from django_dynamic_fixture import G

from ereports.engine.columns import Column, DecimalColumn, RowValue, RowValueError
from ereports.engine.datasource import Datasource
from ereports.tests.engine.fixtures import FakeQuerySet
from ereports.engine.report import BaseReport
from ereports.templatetags.ereports import isNumeric, col_to_css_class, with_widget, format_raw_value, \
    subtotal, total, div_mod, humanize_filters
from ereports.tests.app.models import SimpleDemoModel


def test_isnumeric():
    assert isNumeric(1) == 1
    assert isNumeric("1") == 1
    assert isNumeric(1.0) == 1
    assert isNumeric("1.0") == 1

    assert not isNumeric("wrong")


def test_col_to_css_class():
    c = Column('user.first_name')
    r = RowValue('Test', column=c)
    assert col_to_css_class(r) == 'ereports_engine_columns_column'

    e = RowValueError('Test')
    assert col_to_css_class(e) == 'Test'

    with pytest.raises(ValueError):
        col_to_css_class('Wrong')

    ds = Datasource.as_datasource(model=User,
                                  queryset=FakeQuerySet(model=User, items=[User(username='username1'),
                                                                           User(username='username2')]))

    TestReport = type('TestReport', (BaseReport,), {'model': User, 'datasource': ds})
    report = TestReport.as_report()

    assert col_to_css_class(report[0].username) == 'ereports_engine_columns_charcolumn'
    assert col_to_css_class(report[0]['username']) == 'ereports_engine_columns_charcolumn'


def test_with_widget():
    c = Column('user.first_name')
    r = RowValue('Test', column=c)
    assert with_widget(r) == 'Test'
    assert with_widget(r, format='foo') == 'Test'

    c = Column('normal', format="%Y-%m-%d")
    r = RowValue(datetime.date(2000, 1, 1), column=c)
    assert with_widget(r, format='xls') == datetime.date(2000, 1, 1)

    e = RowValueError('Test')
    assert with_widget(e) == 'Test'

    with pytest.raises(ValueError):
        with_widget('Wrong')


def test_format_raw_value():
    c = DecimalColumn('salary.currency')
    r = RowValue(Decimal("1000"), column=c)
    assert format_raw_value(r) == "1,000.00"

    r = RowValue("1000", column=c)
    assert format_raw_value(r) == "1000"

    ds = Datasource.as_datasource(model=User,
                                  queryset=FakeQuerySet(model=User, items=[User(username='username1'),
                                                                           User(username='username2')]))

    TestReport = type('TestReport', (BaseReport,), {'model': User, 'datasource': ds})
    report = TestReport.as_report()

    assert format_raw_value(report[0].username) == 'username1'
    assert format_raw_value(report[0]['username']) == 'username1'


@pytest.mark.django_db
def test_subtotal():
    G(SimpleDemoModel, n=2, char='abc', integer1=10, integer2=20)
    ds = Datasource.as_datasource(model=SimpleDemoModel)
    r = BaseReport.as_report(datasource=ds)

    assert subtotal(r, r, 'integer1') == 20
    assert subtotal(r, int, 'integer1') == 0


@pytest.mark.django_db
def test_total():
    G(SimpleDemoModel, n=2, char='abc', integer1=10, integer2=20)
    ds = Datasource.as_datasource(model=SimpleDemoModel)
    r = BaseReport.as_report(datasource=ds)

    assert total(r, 'integer2') == 40
    assert total(r, 'char') == 0


def test_div_mod():
    assert div_mod(5, 2) == divmod(5, 2)


def test_humanize_filters():
    f = {
        'eod__lt': datetime.date(2013, 5, 31),
        'nte__gt': datetime.date(2012, 11, 1),
        'eod__gt': datetime.date(2011, 9, 1),
        u'office': 'Testing',
        'admin_duty_station': 'Bangkok'
    }
    r = humanize_filters(f)
    assert r == [
        ('Admin Duty Station', 'Bangkok'),
        ('Eod Gt', '2011-09-01'),
        ('Eod Lt', '2013-05-31'),
        ('Nte Gt', '2012-11-01'),
    ]



# -*- coding: utf-8 -*-
from django.contrib.auth.models import User, Permission
from django.contrib.contenttypes.models import ContentType
from django.db.models import Q
from django.test.testcases import TestCase

from ereports.engine.utils import get_tables_for_query


class TestGetTablesForQuery(TestCase):
    def test_get_tables(self):
        qs = User.objects.filter(username='*')
        tables = get_tables_for_query(qs.query)
        self.assertEqual(tables, ['auth_user'])

    def test_get_tables_children(self):
        qs = Permission.objects.filter(name='*', content_type=ContentType.objects.filter(name='*'))
        tables = get_tables_for_query(qs.query)
        self.assertEqual(sorted(tables), ['auth_permission', 'django_content_type'])

    def test_get_tables_children_wherenode(self):
        qs = Permission.objects.filter(Q(name__startswith='a'), Q(content_type__name='*'))
        tables = get_tables_for_query(qs.query)
        self.assertEqual(sorted(tables), ['auth_permission', 'django_content_type'])



from django.core.exceptions import ValidationError
from django.test.testcases import TestCase
from ereports.models import ReportConfiguration, ReportGroup, ReportTemplate, validate_report_class


class TestReportGroup(TestCase):
    def test_init(self):
        r = ReportGroup(name="Name")
        self.assertEqual(unicode(r), u"Name")


class TestReportTemplate(TestCase):
    def test_init(self):
        r = ReportTemplate(name="Name", body="Body Text", system=True)
        self.assertEqual(unicode(r), u"Name")


class TestValidateReportClass(TestCase):
    def test_validate_report_class_exception(self):
        with self.assertRaises(ValidationError):
            validate_report_class("WrongReport")

    # def test_validate_report_class(self):
    #     validate_report_class(fqn(Payroll))


class TestReportConfiguration(TestCase):
    def test_unicode(self):
        r = ReportConfiguration(name="Name")
        self.assertEqual(unicode(r), u"Name")

    def test_title(self):
        r = ReportConfiguration(name="Name")
        self.assertEqual(r.title(), u"Name")

    def test_get_hard_filters(self):
        r = ReportConfiguration(
            filtering="office={{selected_office}}\nnte__lt={{ today }}\nuser.first_name={{ first_name }}"
        )
        context = dict(
            selected_office="Testing",
            today="Now",
            first_name="User"
        )
        filters = r.get_hard_filters(context)
        self.assertIn('office', filters.keys())
        self.assertIn('nte__lt', filters.keys())
        self.assertIn('user__first_name', filters.keys())
        self.assertEqual(filters['office'], 'Testing')
        self.assertEqual(filters['nte__lt'], 'Now')
        self.assertEqual(filters['user__first_name'], 'User')

    def test_hard_filters_none(self):
        r = ReportConfiguration(
            filtering=None
        )
        context = dict()
        filters = r.get_hard_filters(context)
        self.assertEqual(filters, {})

    def test_get_allowed_group_by(self):
        r = ReportConfiguration(
            groupby='user.first_name'
        )
        g = r.get_allowed_group_by()
        self.assertEqual([('user.first_name', 'user.first_name')], g)

        r = ReportConfiguration(
            groupby='user.first_name  ;  First Name'
        )
        g = r.get_allowed_group_by()
        self.assertEqual([('user.first_name', 'First Name')], g)

        r = ReportConfiguration(
            groupby='''user.first_name;First Name
            user.last_name;Last Name
            random;extra;items
            single
            '''
        )
        g = r.get_allowed_group_by()
        self.assertEqual([('user.first_name', 'First Name'),
                          ('user.last_name', 'Last Name'),
                          ('random', 'extra;items'),
                          ('single', 'single')], g)

    def test_get_allowed_order_by(self):
        r = ReportConfiguration(
            ordering='user.first_name'
        )
        g = r.get_allowed_order_by()
        self.assertEqual([('user.first_name', 'user.first_name')], g)

        r = ReportConfiguration(
            ordering='user.first_name  ;   First Name'
        )
        g = r.get_allowed_order_by()
        self.assertEqual([('user.first_name', 'First Name')], g)

        r = ReportConfiguration(
            ordering='''user.first_name;First Name
            user.last_name;Last Name
            random;extra;items
            single
            '''
        )
        g = r.get_allowed_order_by()
        self.assertEqual([('user.first_name', 'First Name'),
                          ('user.last_name', 'Last Name'),
                          ('random', 'extra;items'),
                          ('single', 'single')], g)

    def test_get_allowed_filters(self):
        r = ReportConfiguration(
            filtering='user.first_name'
        )
        g = r.get_allowed_filters()
        self.assertEqual(['user.first_name'], g)

        r = ReportConfiguration(
            filtering='user=first_name'
        )
        g = r.get_allowed_filters()
        self.assertEqual([], g)

        r = ReportConfiguration(
            filtering='''user.first_name
            user.last_name
            user=email
            '''
        )
        g = r.get_allowed_filters()
        self.assertEqual(['user.first_name', 'user.last_name'], g)



try:
    from django.conf.urls import patterns, url, include
except ImportError: #1.3 fallback
    from django.conf.urls.defaults import *

from django.contrib import admin
admin.autodiscover()

urlpatterns = patterns('',
    (r'^admin/', include(admin.site.urls)),
)




import hashlib
from operator import itemgetter
from django.core.cache import cache as _cache
from django.db.models.signals import post_save, post_delete
from ereports.utils import fqn, flatten


EREPORTS_CACHE_MODELS_PREFIX = 'ereports/models/%s'


def invalidate(sender, **kwargs):
    try:
        _cache.incr(EREPORTS_CACHE_MODELS_PREFIX % fqn(sender), 1)
    except ValueError:
        reset(sender)


def reset(model, **kwargs):
    _cache.set(EREPORTS_CACHE_MODELS_PREFIX % fqn(model), 0)


def monitor_model(model):
    version = _cache.get(EREPORTS_CACHE_MODELS_PREFIX % fqn(model))
    if version is None:
        reset(model)
    post_save.connect(invalidate, sender=model, dispatch_uid="ereports_save_%s" % str(model))
    post_delete.connect(invalidate, sender=model, dispatch_uid="ereports_delete_%s" % str(model))


class BaseCacheManager(object):
    def store(self, key, value, timeout=None, version=None):
        _cache.set(key, value, timeout, version)

    def retrieve(self, cache_key, default=None, version=None):
        return _cache.get(cache_key, default, version)

    def delete(self, cache_key, version=None):
        return _cache.delete(cache_key, version)

    def get_last_cache_version(self, model):
        return _cache.get(EREPORTS_CACHE_MODELS_PREFIX % fqn(model), 0)

    def get_model_gen_part(self, model):
        """
            returns a tuple composed by 'generation prefix', 'generation value'
            fro the passed model

        :param model:
        :return:
        """
        prefix = fqn(model)
        return prefix, self.get_last_cache_version(model)

    def get_key(self, target):
        raise NotImplementedError()


class DummyCacheManager(BaseCacheManager):
    def store(self, key, value, timeout=None, version=None):
        pass

    def retrieve(self, cache_key, default=None, version=None):
        return None

    def delete(self, cache_key, version=None):
        return None

    def get_key(self, target):
        return ''


class DatasourceCacheManager(BaseCacheManager):
    def get_key(self, datasource):
        parts = [self.get_model_gen_part(datasource.model)]

        if datasource.dependent_models:
            for model in datasource.dependent_models:
                parts.append(self.get_model_gen_part(model))

        parts = sorted(parts, key=itemgetter(0))
        parts.insert(0, ('ereports', _cache.get('ereports', 1)))  # to invalidate whole ereports cache

        m = hashlib.md5()
        m.update("{0}-{1}-{2}-{3}".format(str(datasource.columns), str(datasource.kwfilters),
                                          str(datasource.filters), str(datasource.order_by)))
        parts.append(m.hexdigest())
        key = "/".join(map(str, flatten(parts)))

        return key



# -*- encoding: utf-8 -*-
from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import ValidationError
from django.forms import ChoiceField
from django_dynamic_fixture import N
from unittest import TestCase

from ereports.forms import MyModelChoiceField, ReportConfigurationForm
from ereports.engine.registry import registry


class TestMyModelChoiceField(TestCase):
    def test_label_from_instance(self):
        c = N(ContentType, app_label='label', name='name')
        m = MyModelChoiceField(queryset=ContentType.objects.order_by('app_label', 'name'))
        self.assertEqual(m.label_from_instance(c), "Label - Name")


class TestReportConfigurationForm(TestCase):
    def test_init(self):
        form = ReportConfigurationForm()
        self.assertIsInstance(form.fields['report_class'], ChoiceField)
        self.assertEqual(form.fields['report_class'].choices, registry.choices())

    def test_clean_groupby(self):
        form = ReportConfigurationForm()

        form.cleaned_data = {'groupby': "one;two"}
        self.assertEqual(form.clean_groupby(), "one;two")

        with self.assertRaises(ValidationError):
            form.cleaned_data = {'groupby': 'one'}
            form.clean_groupby()



from django.utils.encoding import smart_unicode
from kiowa.utils import get_attr


class Column(list):
    pass


class PayslipGroup(list):

    def sum(self, fieldname):
        tot = 0
        for el in self:
            tot += (get_attr(el, fieldname) or 0)
        return tot


class Regrouper(dict):
    def get(self, k, d=None):
        if not k in self:
            self[k] = PayslipGroup()
        return super(Regrouper, self).get(k, d)

    def __getitem__(self, y):
        if not y in self:
            self[y] = PayslipGroup()
        return super(Regrouper, self).__getitem__(y)

    def values(self):
        return super(Regrouper, self).values()

    def keys(self):
        return super(Regrouper, self).keys()
    groups = keys

    def items(self):
        for k, v in dict.items(self):
            v.sort(key=lambda x: x.contract.employee.index_no)

        sorted = zip(self.keys(), self.values())
        sorted.sort(key=lambda x: smart_unicode(x[0]))
        abs_order = 1
        for group, ps in sorted:
            for payslip in ps:
                payslip.abs_order = abs_order
                abs_order += 1

        return sorted


class Totalizer(dict):
    def __init__(self):
        self.columns = {}

    def get(self, k, d=None):
        if not k in self:
            self[k] = Column()
        return super(Totalizer, self).get(k, d)

    def __getitem__(self, y):
        if not y in self:
            self[y] = Column()
        return super(Totalizer, self).__getitem__(y)



# -*- coding: utf-8 -*-
from decimal import Decimal
from unittest import TestCase

from ereports.engine.columns import Column, normalize_name, get_column_for_attribute, CharColumn, \
    IntegerColumn, DateColumn, DecimalColumn, RowValue, BooleanColumn, CalcColumn, ColumnCallable, OptionalColumn, \
    StringFormatColumn
from ereports.tests.app.models import DemoModel, DemoModelGroup, SimpleDemoModel
from ereports.tests.app.reports import SimpleDemoModelSource


class FakeCallable():
    title = 'Callable'

    def __init__(self, obj, ds):
        if ds is not None:
            raise Exception


def F(obj=None):
    if obj is None:
        return 'called without obj'
    return 'called with obj'


def test_normalize():
    attr_norm_list = [
        ("char", "char"),
        (".hello", "hello"),
        ("one.two..three", "one_two__three"),
        ("1_2.3.hello.world", "hello_world"),
    ]

    for attr, norm in attr_norm_list:
        n = normalize_name(attr)
        assert n == norm


def test_get_column_for_attribute():
    c = get_column_for_attribute(DemoModel, 'char')
    assert c == CharColumn

    c = get_column_for_attribute(DemoModel, 'integer')
    assert c == IntegerColumn

    c = get_column_for_attribute(DemoModel, 'date')
    assert c == DateColumn

    c = get_column_for_attribute(DemoModel, 'decimal')
    assert c == DecimalColumn

    c = get_column_for_attribute(DemoModel, 'logic')
    assert c == BooleanColumn


class TestColumns(TestCase):
    def test_init(self):
        attr = 'char'
        c = Column(attr)
        self.assertEqual(c.attr, attr)
        self.assertEqual(str(c), attr)
        self.assertEqual(repr(c), "'%s'" % attr)

        c = Column(attr, widget='Widget')
        self.assertEqual(c.widget, 'Widget')

    def test_title(self):
        title = 'AAA'
        c = Column('char', title=title)
        self.assertEqual(c.title, title)

        c = Column('char')
        self.assertEqual(c.title, "Char")

        c = Column('char', title="")
        self.assertEqual(c.title, "")

        c = Column('char', model="AAA")
        self.assertEqual(c.title, "Char")

        c = Column('char', model="AAA", title="")
        self.assertEqual(c.title, "")

        c = Column('name', model=DemoModel)
        self.assertEqual(c.title, 'Name')

        c = Column('user.first_name', model=DemoModelGroup)
        self.assertEqual(c.title, 'first name')

    def test_get_value(self):
        o = SimpleDemoModel()
        ds = SimpleDemoModelSource.as_datasource()

        # model found return attr
        c = Column('char')
        v = c.get_value(o, ds)
        self.assertIsInstance(v, RowValue)

        # raise last ValueError
        c = Column('nonexistant')
        with self.assertRaises(ValueError):
            c.get_value(o, ds)

        # Exception with raise ValueError
        c = Column(FakeCallable, name='fake')
        with self.assertRaises(ValueError):
            c.get_value(o, ds)

    def test_get_value_model_attr_callable(self):
        o = SimpleDemoModel()
        ds = SimpleDemoModelSource.as_datasource()
        o.fake = F
        c = Column('fake')
        v = c.get_value(o, ds)
        self.assertIsInstance(v, RowValue)
        self.assertEqual(v, 'called without obj')

    def test_get_value_attr_on_ds(self):
        """Test get value where attr not found on model, but on ds"""
        o = SimpleDemoModel()
        ds = SimpleDemoModelSource.as_datasource()

        ds.fake = 'Hello'
        c = Column('fake')
        v = c.get_value(o, ds)
        self.assertIsInstance(v, RowValue)
        self.assertEqual(v, 'Hello')

    def test_get_value_attr_callable_on_ds(self):
        """Test get value where attr is not found on model and is callable
        attr on ds
        """
        o = SimpleDemoModel()
        ds = SimpleDemoModelSource.as_datasource()

        ds.fake = F
        c = Column('fake')
        v = c.get_value(o, ds)
        self.assertIsInstance(v, RowValue)
        self.assertEqual(v, 'called with obj')

    def test_apply_manipulator(self):
        class FakeCell():
            value = 'low'

        def upper(v):
            return v.upper()

        c = Column('char', manipulator=upper)
        v = c.apply_manipulator(FakeCell())
        self.assertEqual(v.value, 'LOW')


class TestRowValue(TestCase):
    def test_init(self):
        value = 'val'
        r = RowValue(value)
        self.assertEqual(r.value, value)
        self.assertEqual(str(r), str(value))
        self.assertEqual(repr(r), repr(value))
        self.assertEqual(unicode(r), unicode(value))
        self.assertEqual(r, r)
        self.assertEqual(r, value)
        self.assertNotEqual(r, 'wrong')

    def test_pickle(self):
        import pickle
        value = 'val'
        c = Column('char')
        r = RowValue(value, column=c)

        p = pickle.dumps(r)
        up = pickle.loads(p)
        self.assertEqual(up.column, None)


class TestCalcColumn(TestCase):
    def test_init(self):
        col = CalcColumn('char')
        self.assertEqual(col.attrs, 'char')

    def test_get_value(self):
        ds = SimpleDemoModelSource.as_datasource()
        col = CalcColumn(['integer1'])

        self.assertEqual(col.get_value(SimpleDemoModel(integer1=5), ds), 5)

        col.initial = 10
        self.assertEqual(col.get_value(SimpleDemoModel(integer1=5), ds), 15)


class TestColumnCallable(TestCase):
    def test_init(self):
        col = ColumnCallable(FakeCallable)
        self.assertEqual(col.title, 'Callable')

    def test_get_value(self):
        col = ColumnCallable(FakeCallable)
        v = col.get_value(None, None)
        self.assertIsInstance(v, RowValue)

    def test_get_value_exception(self):
        o = SimpleDemoModel()
        ds = SimpleDemoModelSource.as_datasource()

        col = ColumnCallable(FakeCallable)
        with self.assertRaises(ValueError):
            col.get_value(o, ds)


class TestStringFormatColumn(TestCase):
    def test_init(self):
        col = StringFormatColumn('simple')
        self.assertEqual(col.format, '{0}')

    def test_get_value(self):
        col = StringFormatColumn('simple')
        self.assertEqual(col.get_value('data', None), RowValue('data'))


class TestDecimalColumn(TestCase):
    def test_get_value(self):
        col = DecimalColumn('integer1')
        ds = SimpleDemoModelSource.as_datasource()
        v = col.get_value(SimpleDemoModel(integer1=5), ds)
        self.assertEqual(v, Decimal(5.00))


class TestOptionalColumn(TestCase):
    def test_get_value(self):
        ds = SimpleDemoModelSource.as_datasource()

        col = OptionalColumn('nonexistant')
        v = col.get_value(SimpleDemoModel(), ds)
        self.assertEqual(v, "")

        col = OptionalColumn('integer1')
        v = col.get_value(SimpleDemoModel(integer1=5), ds)
        self.assertEqual(v, 5)



# -*- coding: utf-8 -*-
from decimal import Decimal
import re
import operator
import logging
import datetime
from django.utils.encoding import smart_str
from ereports.engine.widgets import ColumnWidget, CurrencyWidget, YesNoWidget, DateWidget, TimeWidget
from django.db import models
from ereports.utils import get_verbose_name, get_field_from_path, get_attr


__all__ = ['Column', 'DecimalColumn', 'DateColumn', 'DateTimeColumn', 'TimeColumn', 'IntegerColumn', 'BooleanColumn']

NOTFOUND = object()
rex = re.compile(r'^(\d|_)+')
logger = logging.getLogger(__name__)


def normalize_name(attr):
    attr = "".join(attr).replace(".", "_")
    return rex.sub("", attr)


class RowValueError(Exception):
    pass


class RowValue(object):
    def __init__(self, value, column=None):
        self.value = value
        self.column = column

    def __str__(self):
        return smart_str(self.value)

    def __repr__(self):
        return repr(self.value)

    def __unicode__(self):
        return unicode(self.value)

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.value == other.value
        return self.value == other

    def __ne__(self, other):
        return not self.__eq__(other)

    def __getstate__(self):
        obj_dict = self.__dict__.copy()
        obj_dict['column'] = None
        return obj_dict


class Column(object):
    widget = ColumnWidget
    wraps = basestring
    default_format = None

    def __init__(self, attr, title=None, manipulator=None, model=None, widget=None, sys_only=False, format=None,
                 name=None, **kwargs):
        self.attr = attr
        self.sys_only = sys_only
        self.name = name or normalize_name(attr)
        self._title = title
        self.format = format or self.default_format
        self._manipulator = manipulator or (lambda x: x)
        self.model = model
        if widget:
            self.widget = widget

    def __repr__(self):
        return repr(self.attr)

    def __str__(self):
        return str(self.attr)

    def get_value(self, obj, datasource):
        value = self._get_value_from_attr(obj, self.attr, datasource)
        return RowValue(value, self)

    def _get_value_from_attr(self, obj, attr_name, datasource):
        try:
            notfound = object()
            attr = get_attr(obj, attr_name, notfound)
            if attr is notfound:
                attr = getattr(datasource, attr_name, None)
                if attr:
                    if callable(attr):
                        return attr(obj)
                    else:
                        return attr
            else:
                if callable(attr):
                    return attr()
                return attr
        except Exception as e:
            logger.exception(e)
            raise ValueError("Unable to get value from `%s`: `%s` `%s`" % (attr_name, type(e), e))
        raise ValueError(attr_name)

    @property
    def title(self):
        if self._title is not None:
            return self._title
        try:
            return get_verbose_name(self.model, self.attr)
        except AttributeError:
            return self.attr.split(".")[-1].replace("_", " ").title()

    def apply_manipulator(self, cell):
        cell.value = self._manipulator(cell.value)
        return cell


class CalcColumn(Column):
    op = operator.add
    initial = 0
    wraps = int

    def __init__(self, attrs, **kwargs):
        self.attrs = attrs
        super(CalcColumn, self).__init__(attrs, **kwargs)

    def get_value(self, obj, datasource):
        result = self.initial
        for el in self.attrs:
            value = self._get_value_from_attr(obj, el, datasource)
            result = self.op(value, result)
        return RowValue(result, self)


class ColumnCallable(Column):
    """
        wraps a callable as Column.

        callable will be invoked passing the current object and the Datasoource instance.
        the Column header can be customized with ``callable.title = 'Title'``

        >>> def get_custom_format(obj, ds):
        ...     return obj.field.upper()
        >>> get_custom_format.title = 'Field'
        >>> c = ColumnCallable(get_custom_format)

    """

    def __init__(self, attr, **kwargs):
        name = kwargs.pop('name', None) or normalize_name(attr.__name__)
        title = kwargs.pop('title', None) or getattr(attr, 'title', name)
        super(ColumnCallable, self).__init__(attr, title=title, name=name, **kwargs)

    def get_value(self, obj, datasource):
        try:
            value = self.attr(obj, datasource)
            return RowValue(value, self)
        except Exception as e:
            logger.exception(e)
            raise ValueError("ColumnCallable: Unable to get value from `%s`: `%s`" % (self.attr, e))


class StringFormatColumn(Column):
    def __init__(self, attr, title=None, manipulator=None, model=None, widget=None, sys_only=False, format='{0}',
                 name=None, **kwargs):
        super(StringFormatColumn, self).__init__(attr, title, manipulator, model, widget, sys_only, format, name,
                                                 **kwargs)

    def get_value(self, obj, datasource):
        result = self.format.format(obj)
        return RowValue(result, self)


class DecimalColumn(Column):
    wraps = Decimal

    def get_value(self, obj, datasource):
        value = self._get_value_from_attr(obj, self.attr, datasource)
        return RowValue(self.wraps(value), self)


class DateColumn(Column):
    wraps = datetime.date
    widget = DateWidget
    default_format = '%d %b %Y'


class DateTimeColumn(Column):
    wraps = datetime.datetime


class TimeColumn(Column):
    wraps = datetime.time
    widget = TimeWidget
    default_format = '%H %M'


class IntegerColumn(Column):
    wraps = int
    pass


class CharColumn(Column):
    pass


class TextColumn(Column):
    pass


class CurrencyColumn(Column):
    widget = CurrencyWidget


class BooleanColumn(Column):
    wraps = bool
    widget = YesNoWidget


class OptionalColumn(Column):
    def get_value(self, obj, datasource):
        try:
            value = self._get_value_from_attr(obj, self.attr, datasource)
        except ValueError as e:
            logger.exception(e)
            value = ""
        return RowValue(value, self)


def get_column_for_attribute(model, fieldpath):
    """
        returns the more suitable Column class for `attr`.

    :param: can be a Django Field, attribute or callable
    :return:
    """
    attr = get_field_from_path(model, fieldpath)
    mapping = {
        models.IntegerField: IntegerColumn,
        models.DecimalField: DecimalColumn,
        models.DateField: DateColumn,
        models.CharField: CharColumn,
        models.BooleanField: BooleanColumn,
    }
    return mapping.get(type(attr), Column)



from unittest import TestCase
from django.contrib.auth.models import User
from ereports.engine.columns import StringFormatColumn
from ereports.engine.datasource import Datasource
from ereports.engine.report import BaseReport
from ereports.tests.engine.fixtures import FakeQuerySet


class TestGroup(TestCase):
    def test_simple_group(self):
        itms = [User(username='username%s' % x, is_staff=bool(x % 2)) for x in range(6)]

        ds = Datasource.as_datasource(model=User,
                                      queryset=FakeQuerySet(model=User, items=itms))

        TestReport = type('TestReport', (BaseReport,), {'model': User,
                                                        'list_display': ('is_staff', 'username'),
                                                        'group_by': ('is_staff', 'username'),
                                                        'datasource': ds})
        report = TestReport.as_report()
        self.assertSequenceEqual(list(report),
                                 [(False, 'username0'),
                                  (False, 'username2'),
                                  (False, 'username4'),
                                  (True, 'username1'),
                                  (True, 'username3'),
                                  (True, 'username5')])

    def test_callback_group(self):
        itms = [User(username='username%s' % x, is_staff=bool(x % 2)) for x in range(6)]

        ds = Datasource.as_datasource(model=User,
                                      queryset=FakeQuerySet(model=User, items=itms))

        def is_staff(rec):
            return str(rec.is_staff)

        TestReport = type('TestReport', (BaseReport,), {'model': User,
                                                        'list_display': ('is_staff', 'username'),
                                                        'group_by': (is_staff, 'username'),
                                                        'datasource': ds})
        report = TestReport.as_report()
        self.assertItemsEqual(dict(report.get_groups()).keys(), ['True', 'False'])
        self.assertSequenceEqual(list(report),
                                 [(False, 'username0'),
                                  (False, 'username2'),
                                  (False, 'username4'),
                                  (True, 'username1'),
                                  (True, 'username3'),
                                  (True, 'username5')])

    def test_by_custom_column(self):
        itms = [User(username='username%s' % x,
                     first_name='fn%s' % int(x % 2),
                     last_name='ln%s' % int(x % 2),
                     is_staff=bool(x % 2)) for x in range(6)]

        ds = Datasource.as_datasource(model=User,
                                      columns=('is_staff',
                                               'username',
                                               StringFormatColumn('group',
                                                                  format='{0.last_name} {0.first_name}')),
                                      queryset=FakeQuerySet(model=User, items=itms))

        TestReport = type('Report', (BaseReport,), {'datasource': ds,
                                                    'group_by': ('group', 'username'),
                                                    'list_display': ('group', 'username')})
        report = TestReport.as_report()
        self.assertSequenceEqual(sorted(dict(report.get_groups()).keys()),
                                 ['ln0 fn0', 'ln1 fn1'])

        self.assertSequenceEqual(list(report),
                                 [('ln0 fn0', 'username0'),
                                  ('ln0 fn0', 'username2'),
                                  ('ln0 fn0', 'username4'),
                                  ('ln1 fn1', 'username1'),
                                  ('ln1 fn1', 'username3'),
                                  ('ln1 fn1', 'username5')])



from django.db.models.query import EmptyQuerySet, QuerySet


class FakeQuerySet(QuerySet):
    """Turn a list into a Django QuerySet... kind of."""

    def __init__(self, model=None, query=None, using=None, items=None):
        super(FakeQuerySet, self).__init__(model, query, using)
        self._result_cache = items or []

    def count(self):
        return len(self)

    def _clone(self, klass=None, setup=False, **kwargs):
        return FakeQuerySet(self.model, items=self._result_cache)



# -*- coding: utf-8 -*-
from collections import defaultdict
from django.core.exceptions import ImproperlyConfigured
from django.utils.encoding import smart_unicode
from ereports.engine.config import ConfigurationForm
from ereports.engine.datasource import Datasource, DatasourceRow
from ereports.engine.renderer import BaseHtmlRender
from ereports.utils import get_attr, fqn


class Group(list):
    pass


class BaseGrouper(object):
    def __init__(self, report, group_by, internal_order):
        super(BaseGrouper, self).__init__()
        self.report = report
        self.group_by = group_by
        self.internal_order = internal_order  # field name to use for group order
        self._dict = defaultdict(Group)
        self._processed = False

    def _process(self):
        if self._processed:
            return
        ds = self.report.datasource
        # record is a DatasourceRow instance
        for datasourcerow in ds:
            if callable(self.group_by):
                group_name = self.group_by(datasourcerow._original)
            elif isinstance(self.group_by, basestring):
                try:
                    col = self.report.get_column_by_name(self.group_by)
                    group_name = col.get_value(datasourcerow._original, self.report.datasource).value
                except KeyError:
                    group_name = get_attr(datasourcerow._original, self.group_by)
            else:
                group_name = get_attr(datasourcerow._original, self.group_by)

            #datasourcerow._sort_func = self.report._order_columns
            orderedrow = DatasourceRow([(name, datasourcerow[name]) for name in self.report.display_order()])
            orderedrow._original = datasourcerow._original
            self._dict[group_name].append(orderedrow)

        self._processed = True

    def values(self):
        self._process()
        return self._dict.values()

    def keys(self):
        self._process()
        return self._dict.keys()

    groups = keys

    def items(self):
        self._process()

        # order internal group elements
        for group_name, rows in self._dict.items():
            rows.sort(key=lambda x: get_attr(x._original, self.internal_order))

        #returns groups ordered
        sorted_groups = zip(self._dict.keys(), self._dict.values())
        sorted_groups.sort(key=lambda x: smart_unicode(x[0]))

        return sorted_groups


class BaseReport(object):
    title = None
    description = None
    datasource = None
    model = None
    config_form_class = ConfigurationForm  # search form base class
    list_display = None  # columns and order to display
    list_filter = ()  # field names to use in the search_form
    formats = [('html', BaseHtmlRender)]
    use_cache = False
    order_by = None

    group_by = None  # (group, internal order)
    grouper = BaseGrouper
    extras = None
    column_totals = None  # columns to sum the values

    def __init__(self, **kwargs):
        self.extras = kwargs.pop('extras', {})
        for key, value in kwargs.iteritems():
            setattr(self, key, value)
        self.title = self.title or fqn(self)

    def __repr__(self):
        return "<%s: on `%s`>" % (self.__class__.__name__, self.datasource.model.__name__)

    def has_subtotals(self):
        return bool(self.column_totals)

    @classmethod
    def as_report(cls, **initkwargs):
        for key, value in initkwargs.items():
            if not hasattr(cls, key) and not callable(value):
                raise TypeError(u"%s() received an invalid keyword %r" % (cls.__name__, key))

        model = initkwargs.get('model', cls.model)
        datasource = initkwargs.get('datasource', cls.datasource)
        use_cache = initkwargs.get('use_cache', cls.use_cache)
        extras = initkwargs.get('extras', cls.extras)

        if datasource is not None:
            initkwargs['model'] = datasource.model
            initkwargs['datasource'] = datasource._clone(extras=extras)
        elif model is not None:
            initkwargs['datasource'] = Datasource.as_datasource(model=model, use_cache=use_cache)
        else:
            raise ImproperlyConfigured(u"%(cls)s is missing a datasource. Define "
                                       u"%(cls)s.model, %(cls)s.datasource" % {'cls': cls.__name__})

        return cls(**initkwargs)

    def display_order(self):
        """
            returns the order and the column names to display
        :return: list
        """
        if self.list_display is not None:
            return self.list_display
        else:
            return [c.name for c in self.datasource.columns if not c.sys_only]

    def _order_columns(self, row):
        """
            orders the columns of the passed row to reflect the `list_display` attribute
        :param row:
        :return:
        """
        if self.list_display:
            return DatasourceRow([(name, row[name]) for name in self.list_display])
        return row

    def __iter__(self):
        self.datasource.order_by = self.order_by
        if self.group_by:
            g = self.get_groups()
            for group, rows in g:
                for row in rows:
                    yield self._order_columns(row)
        else:
            for row in self.datasource:
                yield self._order_columns(row)

    def __getitem__(self, item):
        rows = self.datasource[item]
        if isinstance(rows, (slice, list, tuple)):
            return tuple([self._order_columns(r) for r in rows])
        return self._order_columns(rows)

    def get_column_by_name(self, name):
        """
            return the Column instance for the given column's name
        :param name: string
        :return:
        """
        return self.columns_dict[name]

    @property
    def columns_dict(self):
        return dict([(c.name, c) for c in self.datasource.columns])

    def get_column_values(self, column_name):
        """
            returns all the values for the given column's name
        :param column_name:
        :return:
        """
        ds = self.datasource
        return [row[column_name].value for row in ds]

    @property
    def headers(self):
        if self.list_display:
            return [self.get_column_by_name(c).title for c in self.list_display
                    if not self.get_column_by_name(c).sys_only]
        else:
            return [c.title for c in self.datasource.columns if not c.sys_only]

    def get_format_labels(self):
        return self.formats_dict.keys()

    @property
    def formats_dict(self):
        return dict(self.formats)

    def get_renderer_class_for_format(self, render_format):
        return self.formats_dict[render_format]

    def get_renderer_for_format(self, render_format):
        return self.formats_dict[render_format](self)

    def get_groups(self):
        if not self.group_by:
            raise ImproperlyConfigured('Cannot use get_group() without set group_by')

        assert len(self.group_by) == 2, "Invalid GroupBy `%s`" % self.group_by
        groups = self.grouper(self, *self.group_by)
        return groups.items()



from django import forms
from django.contrib import admin
from django.db import models
from ereports.forms import ReportConfigurationForm
from ereports.models import ReportConfiguration, ReportGroup, ReportTemplate


class IReport(admin.ModelAdmin):
    save_on_top = True
    list_display = ('name', 'description', 'target_model', 'report_class', 'published', 'template', 'group')
    list_filter = ('published', 'template', 'group')
    form = ReportConfigurationForm
    actions = ['publish', 'unpublish']

    def publish(self, request, queryset):
        queryset.update(published=True)

    def unpublish(self, request, queryset):
        queryset.update(published=False)

    formfield_overrides = {
        models.TextField: {'widget': forms.Textarea({'cols': '60', 'rows': '20'})}, }
    #
    search_fields = ('name', )
    fieldsets = [
        (None, {'fields': (('name', 'description', 'target_model', 'published', 'select_related', 'use_distinct'),
                           ('template', 'group', 'report_class'))}),
        ('Columns', {'classes': ['collapse', 'textarea-fields', ],
                     'fields': [('columns', 'filtering', 'ordering', 'groupby'), ]}),
    ]


admin.site.register(ReportConfiguration, IReport)
admin.site.register(ReportTemplate, admin.ModelAdmin)
admin.site.register(ReportGroup, admin.ModelAdmin)



# -*- coding: utf-8 -*-
from django.contrib.auth.models import User
from django.http import QueryDict
from django.test.client import RequestFactory
from django_dynamic_fixture import G


def get_fake_request():
    u = G(User, username='sax', password='123')
    setattr(u, 'is_authenticated()', True)
    setattr(u, 'selected_office', False)

    request = RequestFactory().request()
    request.user = u

    querydict = QueryDict('arg2=one')
    querydict = querydict.copy()
    querydict.update({'arg1': 'test', 'arg2': "two"})
    request.GET = querydict
    request.POST = QueryDict('username=random')

    return request



# -*- coding: utf-8 -*-
from django.http import HttpResponse
from django_dynamic_fixture import G
from django_webtest import WebTest

from ereports.engine.renderer import BaseHtmlRender, BaseXlsRender
from ereports.tests.app.models import SimpleDemoModel
from ereports.tests.app.reports import SimpleDemoReport, SimpleDateReport
from ereports.tests.fixtures import get_fake_request


class TestBaseHtmlRenderer(WebTest):
    def test_render_to_response(self):
        report = SimpleDemoReport.as_report()
        renderer = report.get_renderer_for_format('html')
        self.assertIsInstance(renderer, BaseHtmlRender)
        request = get_fake_request()
        r = renderer.render_to_response(request)
        self.assertIsInstance(r, HttpResponse)

    def test_render_to_response_with_groupby(self):
        report = SimpleDateReport.as_report()
        renderer = report.get_renderer_for_format('html')
        self.assertIsInstance(renderer, BaseHtmlRender)
        request = get_fake_request()
        r = renderer.render_to_response(request)
        self.assertIsInstance(r, HttpResponse)


class TestBaseXlsRenderer(WebTest):
    def test_render_to_response(self):
        G(SimpleDemoModel, n=2, char='abc', integer1=1, integer2=2, boolean=True)
        G(SimpleDemoModel, n=2, char=u'êbc', integer1=1, integer2=2, boolean=True)
        report = SimpleDemoReport.as_report()
        renderer = report.get_renderer_for_format('xls')
        self.assertIsInstance(renderer, BaseXlsRender)
        request = get_fake_request()
        r = renderer.render_to_response(request)
        self.assertIsInstance(r, HttpResponse)



def get_tables_for_query(query):
# from johnny-cache
# https://github.com/jmoiron/johnny-cache/blob/master/johnny/cache.py

    """
Takes a Django 'query' object and returns all tables that will be used in
that query as a list. Note that where clauses can have their own
querysets with their own dependent queries, etc.
"""
    from django.db.models.sql.where import WhereNode
    from django.db.models.query import QuerySet

    tables = [v[0] for v in getattr(query, 'alias_map', {}).values()]

    def get_tables(node, tables):
        for child in node.children:
            if isinstance(child, WhereNode):  # and child.children:
                tables = get_tables(child, tables)
            elif not hasattr(child, '__iter__'):
                continue
            else:
                for item in (c for c in child if isinstance(c, QuerySet)):
                    tables += get_tables_for_query(item.query)
        return tables

    if query.where and query.where.children:
        where_nodes = [c for c in query.where.children if isinstance(c, WhereNode)]
        for node in where_nodes:
            tables += get_tables(node, tables)

    return list(set(tables))



from django_webtest import WebTest
from ereports.engine.mixins import ReportRender
from ereports.tests.app.reports import SimpleDemoReport


class TestReportRender(WebTest):
    def test_init(self):
        r = ReportRender(SimpleDemoReport)
        self.assertEqual(r.report, SimpleDemoReport)



# -*- coding: utf-8 -*-
import pytest

from ereports.template_loaders import EReportTemplateLoader
from django.template import TemplateDoesNotExist


@pytest.mark.django_db
def test_load_template_source():
    template_name = 'ereports/tpls/report.html'
    t = EReportTemplateLoader()

    with pytest.raises(TemplateDoesNotExist):
        t.load_template_source(template_name)



# -*- encoding: utf-8 -*-

from django.conf.urls import patterns, url
from pasportng.pasport.urls2 import secure_url
from .views import ReportIndex, PayrollView, ReportView, ReportConfigView


urlpatterns = patterns(
    '',

    secure_url(r'^(?P<office>[\w\d]+)/$',
               ReportIndex.as_view(),
               name='ereports.list_office_reports'),

    secure_url(r'^(?P<office>[\w\d]+)/(?P<year>\d+)/(?P<month>\w+)/(?P<format>\w+)/$',
               PayrollView.as_view(),
               name="ereports.payroll_of"),

    secure_url(r'^(?P<office>[\w\d]+)/report/(?P<report>\d+)/configure/$',
               ReportConfigView.as_view(),
               name='ereports.configure'),

    secure_url(r'^(?P<office>[\w\d]+)/(?P<report>\d+)/(?P<format>\w+)/$',
               ReportView.as_view(),
               name='ereports.do_report'),

    secure_url(r'^(?P<office>[\w\d]+)/(?P<report>\d+)/(?P<format>(html|csv|xls))/$',
               ReportView.as_view(),
               name='ereports.do_report'),

    secure_url(r'^(?P<office>[\w\d]+)/(?P<report>\d+)/$',
               ReportView.as_view(), name='ereports.do_report',
               kwargs={'format': 'html'}),

    url(r'^model/info/(?P<owner>.+)/(?P<app_name>\w+)/(?P<model_name>\w+)/$', 'pasportng.ereports.views.get_model_meta',
        name='get_model_meta'),

)



# -*- coding: utf-8 -*-
import datetime

from django.contrib.contenttypes.models import ContentType
from django.http import QueryDict
from django_dynamic_fixture import G
from django_webtest import WebTest
from mock import patch

from ereports.views import ReportIndex, ReportFilter
from ereports.models import ReportConfiguration
from ereports.tests.app.models import SimpleDemoModel

from .fixtures import get_fake_request


class TestReportIndex(WebTest):
    def setUp(self):
        self.request = get_fake_request()

    def test_get_no_reports(self):
        response = ReportIndex.as_view()(self.request)
        self.assertIn('reports', response.context_data)
        self.assertEqual(response.context_data['reports'], [])

    def test_get(self):
        G(ReportConfiguration, name='Standard', published=True)
        response = ReportIndex.as_view()(self.request)
        self.assertIn('reports', response.context_data)
        self.assertEqual(len(response.context_data['reports']), 1)


class TestReportFilter(WebTest):
    def setUp(self):
        self.request = get_fake_request()
        self.config = G(ReportConfiguration,
                        name='Standard',
                        target_model=ContentType.objects.get_for_model(SimpleDemoModel),
                        report_class='ereports.tests.app.reports.SimpleDemoReport',
                        filtering='',
                        ordering='',
                        groupby='')
        self.querydict = QueryDict('')
        self.querydict = self.querydict.copy()
        self.querydict.update({'_format': 'html',
                               '_report_list_display': 'integer1'})

    def test_get(self):
        response = ReportFilter.as_view()(self.request, pk=self.config.pk)
        self.assertIn('configuration', response.context_data)
        self.assertIn('today', response.context_data)
        self.assertEqual(response.context_data['today'].strftime("%Y-%m-%d"),
                         datetime.datetime.today().strftime("%Y-%m-%d"))

        self.assertIn('user', response.context_data)
        self.assertEqual(response.context_data['user'], self.request.user)

    def test_post_invalid(self):
        self.request.method = 'POST'
        #TODO catch message error value and confirm
        with patch('ereports.views.messages'):
            response = ReportFilter.as_view()(self.request, pk=self.config.pk)
            self.assertIn('form', response.context_data)

    def test_post_valid_no_ds(self):
        self.request.method = 'POST'
        self.request.POST = self.querydict
        #TODO catch message error value and confirm
        with patch('ereports.views.messages'):
            response = ReportFilter.as_view()(self.request, pk=self.config.pk)
            self.assertIn('query', response.context_data)

    def test_post_valid(self):
        self.request.method = 'POST'
        self.request.POST = self.querydict
        G(SimpleDemoModel, char='abc', integer1=10)
        response = ReportFilter.as_view()(self.request, pk=self.config.pk)
        self.assertIn('ereports.engine.report.BaseReport', str(response))
        self.assertIn('Integer #1', str(response))



# -*- coding: utf-8 -*-
from unittest import TestCase
from ereports.engine.exceptions import FilterDefinitionError


class TestFilterDefinitionError(TestCase):
    def test_init(self):
        e = FilterDefinitionError('oops')
        self.assertEqual(e.message, 'oops')
        self.assertEqual(str(e), 'oops')
        self.assertEqual(repr(e), 'oops')



#!/usr/bin/env python
import os
import codecs
from setuptools import setup, find_packages
import ereports

RELEASE = ereports.get_version()

base_url = 'https://github.com/saxix/django-concurrency/'
VERSIONMAP = {'final': (ereports.VERSION, 'Development Status :: 5 - Production/Stable'),
              'rc': (ereports.VERSION, 'Development Status :: 4 - Beta'),
              'beta': (ereports.VERSION, 'Development Status :: 4 - Beta'),
              'alpha': ('master', 'Development Status :: 3 - Alpha')}

download_tag, development_status = VERSIONMAP[ereports.VERSION[3]]


def read(*parts):
    here = os.path.abspath(os.path.dirname(__file__))
    return codecs.open(os.path.join(here, *parts), 'r').read()


def reqs(*parts):
    here = os.path.abspath(os.path.dirname(__file__))
    return filter(lambda x: not x.startswith('#'),
                  codecs.open(os.path.join(here, *parts), 'r').readlines())


setup(
    name=ereports.NAME,
    version=RELEASE,
    url='https://github.com/saxix/django-easy-reports',
    author='sax',
    author_email='s.apostolico@gmail.com',
    packages=find_packages(),
    include_package_data=True,

    dependency_links=['http://pypi.wfp.org/simple/'],
    install_requires=reqs('ereports/requirements/install.pip'),
    tests_require=reqs('ereports/requirements/testing.pip'),
    test_suite='conftest.runtests',

    description="Django reporting library ",
    long_description=open('README.rst').read(),
    license="MIT License",
    keywords="django",
    classifiers=[
        development_status,
        'Environment :: Web Environment',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Software Development :: Libraries :: Application Frameworks',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
    platforms=['any'],

)



# -*- coding: utf-8 -*-
from ereports.engine import registry
from ereports.utils import fqn

from ereports.tests.app.reports import SimpleDemoReport


class DemoReport(object):
    pass


def test_init():
    r = registry.Registry()
    assert r.reports == []


def test_register():
    r = registry.Registry()
    r.register(DemoReport)
    assert len(r.reports) == 1


def test_register_monitor():
    r = registry.Registry()
    r.register(SimpleDemoReport)
    assert len(r.reports) == 1
    assert len(r.monitors) == 1


def test_get():
    r = registry.Registry()
    r.register(DemoReport)
    assert r.get(fqn(DemoReport), None) == DemoReport
    assert r.get("Wrong", None) is None
    assert r.get("Wrong", DemoReport) == DemoReport
    assert r[0] == r.reports[0]
    assert r[fqn(DemoReport)] == DemoReport
    assert r.choices() == [(fqn(DemoReport), fqn(DemoReport))]



import hashlib
import logging
import datetime
from django.contrib import messages
from django.core.cache import cache
from django.views.generic import TemplateView
from ereports.engine.config import reportform_factory
from ereports.engine.registry import registry
from ereports.engine.report import BaseReport
from ereports.filtering import FilterQuerysetMixin
from ereports.models import ReportConfiguration

logger = logging.getLogger(__name__)


class ReportIndex(TemplateView):
    """ Index of all available reports for an office
    """
    template_name = 'ereports/index.html'

    def get_context_data(self, **kwargs):
        reports = []
        for report in ReportConfiguration.objects.filter(published=True).order_by('group', 'name'):
            reports.append(report)
        return {'reports': reports}


class ReportFilter(FilterQuerysetMixin, TemplateView):
    """ Filter and generated selected report
    """
    template_name = 'ereports/filter.html'
    method = 'POST'

    def get_config(self):
        return ReportConfiguration.objects.get(pk=self.kwargs.get('pk'))

    def get_report(self, **kwargs):
        if self.config.target_model:
            kwargs['model'] = self.config.target_model.model_class()

        Class = self.get_report_class()
        return Class.as_report(**kwargs)

    def get_report_class(self):
        return registry.get(self.config.report_class, BaseReport)

    def get_search_form_class(self):
        kwargs = {'order_by': self.config.get_allowed_order_by(),
                  'filtering': self.config.get_allowed_filters(),
                  'groupby': self.config.get_allowed_group_by()}
        return reportform_factory(self.report, bases=(self.report.config_form_class,), **kwargs)

    def get_search_form(self, *args, **kwargs):
        return self.get_search_form_class()(*args, **kwargs)

    def get_context_data(self, **kwargs):
        today = datetime.datetime.today()

        kwargs.update({
            'configuration': self.config,
            'today': today,
            'user': self.request.user,
        })
        return super(ReportFilter, self).get_context_data(**kwargs)

    def get(self, request, *args, **kwargs):
        self.config = self.get_config()
        self.report = self.get_report()
        kwargs.setdefault('form', self.get_search_form())
        return super(ReportFilter, self).get(request, *args, **kwargs)

    def _run_report(self, request, *args, **kwargs):
        pass

    def post(self, request, *args, **kwargs):
        self.config = self.get_config()
        self.report = self.get_report()

        form = self.get_search_form(request.POST)
        context = self.get_context_data()
        if form.is_valid():
            filters, kwfilters = form.get_filters()
            kwfilters.update(**self.config.get_hard_filters(context))

            report_attributes = form.get_report_attributes()
            self.report = self.get_report(**report_attributes)

            self.report.datasource.add_filters(*filters, **kwfilters)

            ds = self.report.datasource
            m = hashlib.md5()
            m.update("filters={0}".format(str(kwfilters)))
            m.update("list_display={0}".format(str(self.report.list_display)))
            m.update("report_attributes={0}".format(str(report_attributes)))

            cache_key = m.hexdigest()
            context['hash'] = cache_key
            page = cache.get(cache_key)
            if page:
                return page

            if ds:
                context['report'] = self.report
                context['filters_legend'] = form.get_filters_summary()
                context['filters'] = kwfilters
                renderer = self.report.get_renderer_for_format(form.cleaned_data['_format'])
                page = renderer.render_to_response(request, **context)

                # cache.set(cache_key, page)
                return page
            else:
                messages.error(self.request, "Report has no data to display")
                kwargs.setdefault('form', form)
                kwargs.setdefault('query', ds.query)
                logger.debug('Report returns no data `%s`' % ds.query)
                return super(ReportFilter, self).get(request, *args, **kwargs)
        else:
            messages.error(self.request, (str(form.errors).replace('\n', ';')))
            return super(ReportFilter, self).get(request, *args, form=form)



# -*- coding: utf-8 -*-
from importlib import import_module
from inspect import isclass
from logging import getLogger
from django.core.exceptions import ImproperlyConfigured
from django.db import models
from django.utils.encoding import smart_str
from six import string_types


logger = getLogger(__name__)


def get_model_field_names(model):
    return [f.name for f in model._meta.fields]


def currencyformat(value, thousand_sep=',', decimal_sep='.'):
    """
    Returns the currency format of the value param using specified separators

    Usage: {% value_to_format|currencyformat %}
    """
    sign = {False: '-', True: ''}[value > 0]
    integer_part_string = str(abs(int(value)))
    f = lambda x, y, list=[]: f(x[:-y], y, [(x[-y:])] + list) if x else list
    integer_part = thousand_sep.join(f(integer_part_string, 3))
    return "%s%s%s%s" % (sign, integer_part, decimal_sep, ("%0.2f" % value)[-2:])


def fqn(o):
    """Returns the fully qualified class name of an object or a class

    >>> import pytest
    >>> with pytest.raises(ValueError):
    ...    fqn('hello')

    >>> fqn(fqn)
    'ereports.utils.function'

    >>> class FakeClass():
    ...    pass
    >>> fqn(FakeClass)
    'ereports.utils.FakeClass'

    :param o: object or class
    :return: class name
    """
    parts = []
    if not hasattr(o, '__module__'):
        raise ValueError('Invalid argument `%s`' % o)
    parts.append(o.__module__)
    if isclass(o):
        parts.append(o.__name__)
    else:
        parts.append(o.__class__.__name__)
    return ".".join(parts)


def flatten(x):
    """flatten(sequence) -> list

    Returns a single, flat list which contains all elements retrieved
    from the sequence and all recursively contained sub-sequences
    (iterables).

    Examples:

    >>> [1, 2, [3,4], (5,6)]
    [1, 2, [3, 4], (5, 6)]

    >>> flatten([[[1,2,3], (42,None)], [4,5], [6], 7, (8,9,10)])
    [1, 2, 3, 42, None, 4, 5, 6, 7, 8, 9, 10]"""

    result = []
    for el in x:
        # if isinstance(el, (list, tuple)):
        if hasattr(el, "__iter__") and not isinstance(el, string_types):
            result.extend(flatten(el))
        else:
            result.append(el)
    return list(result)


def get_verbose_name(model_or_queryset, field_path):
    """
        returns the verbose name of a Field.

        Examples:

        >>> import pytest
        >>> class Fake(models.Model):
        ...     user = models.CharField(verbose_name='user', max_length=30)
        >>> get_verbose_name(Fake(), 'user')
        'user'
        >>> get_verbose_name(Fake, 'user')
        'user'
        >>> with pytest.raises(AttributeError):
        ...     get_verbose_name(Fake, 'wrong') # doctest: +IGNORE_EXCEPTION_DETAIL, +ELLIPSIS

    :param model_or_queryset: can be a Model/Manager or Queryset
    :param field_path: dotted path to rhe field name
    :return: verbose_name attrinute of Field
    """

    model = getattr(model_or_queryset, 'model', model_or_queryset)
    field = get_field_from_path(model, field_path)
    if field and hasattr(field, 'verbose_name'):
        label = field.verbose_name
    elif field:
        label = field.name.replace('.', ' ').replace('_', ' ')
    else:
        raise AttributeError(field_path)
    return smart_str(label)


def get_field_from_path(Model, field_path):
    """
    get a Model and a path to a attribute, return the field object

    a = get_field_from_path(Model, 'fk.fk.fieldname')
    """
    parts = field_path.split('.')
    target = parts[0]
    if target in Model._meta.get_all_field_names():
        field_object, __, __, __ = Model._meta.get_field_by_name(target)
        if isinstance(field_object, models.fields.related.ForeignKey):
            if parts[1:]:
                return get_field_from_path(field_object.rel.to, '.'.join(parts[1:]))
            else:
                return field_object
        else:
            return field_object
    return None


def get_attr(obj, attr, default=None):
    """Recursive get object's attribute. May use dot notation.

    >>> class C(object): pass
    >>> a = C()
    >>> a.b = C()
    >>> a.b.c = 4
    >>> get_attr(a, 'b.c')
    4

    >>> get_attr(a, 'b.c.y', None)

    >>> get_attr(a, 'b.c.y', 1)
    1
    """
    if '.' not in attr:
        return getattr(obj, attr, default)
    else:
        L = attr.split('.')
        return get_attr(getattr(obj, L[0], default), '.'.join(L[1:]), default)


class StartWithList(list):
    """List redefined to check list items start with particular string

    >>> l = StartWithList(['123', '231', '321'])
    >>> '1' in l
    True
    >>> '23' in l
    True
    >>> '321' in l
    True
    >>> '21' in l
    False
    >>> '4' in l
    False
    """
    def __contains__(self, y):
        for el in self:
            if el.startswith(y):
                return True
        return False


def filter_dictionary(dictionary, filter_func):
    """
        returns the first element of `dictionary` where the element's key pass the filter_func.

        filter_func can be either a callable or a value.
        - if callable filtering is checked with `test(element_value)`
        - if value filtering is checked with `element_value == filter_func`

    :param dictionary:
    :param test:
    :return:

    >>> filter_dictionary({'arg': 'test'}, 'test')
    'arg'
    >>> filter_dictionary({}, 'test')

    >>> def is_test(value):
    ...    return value == 'test'
    >>> filter_dictionary({'arg': 'test'}, is_test)
    'arg'
    """
    if not callable(filter_func):
        test_func = lambda x: x == filter_func
    else:
        test_func = filter_func
    for key, value in dictionary.iteritems():
        if test_func(value):
            return key


def import_by_path(dotted_path, error_prefix=''):
    """
    Import a dotted module path and return the attribute/class designated by the
    last name in the path. Raise ImproperlyConfigured if something goes wrong.

    >>> import pytest
    >>> from django.contrib.auth.models import User
    >>> from django.core.exceptions import ImproperlyConfigured
    >>> with pytest.raises(ImproperlyConfigured):
    ...    import_by_path("wrongpath")

    >>> with pytest.raises(ImproperlyConfigured):
    ...    import_by_path("wrong.path")

    >>> with pytest.raises(ImproperlyConfigured):
    ...    import_by_path("django.contrib.auth.models.Wrong")

    >>> cls = import_by_path("django.contrib.auth.models.User")
    >>> assert isinstance(cls(), User)
    """
    try:
        module_path, class_name = dotted_path.rsplit('.', 1)
    except ValueError:
        raise ImproperlyConfigured("%s%s doesn't look like a module path" % (
            error_prefix, dotted_path))
    try:
        module = import_module(module_path)
    except ImportError as e:
        raise ImproperlyConfigured('%sError importing module %s: "%s"' % (
            error_prefix, module_path, e))
    try:
        attr = getattr(module, class_name)
    except AttributeError:
        raise ImproperlyConfigured('%sModule "%s" does not define a "%s" attribute/class' % (
            error_prefix, module_path, class_name))
    return attr



# -*- coding: utf-8 -*-
from logging import getLogger
import operator
from django.core.exceptions import ImproperlyConfigured
from django.db import models
from django import forms
from ereports.utils import StartWithList, get_field_from_path, filter_dictionary

logger = getLogger(__name__)


class QueryField(forms.CharField):
    def __init__(self, max_length=None, min_length=None, label="", required=False,
                 widget=forms.TextInput({"class": "search-query",
                                         "placeholder": "Type something…"}), *args, **kwargs):
        super(QueryField, self).__init__(max_length, min_length, label=label, required=required,
                                         widget=widget, *args, **kwargs)


class SearchForm(forms.Form):
    pass


ANY_LOOKUP_ALLOWED = 888


class FilterQuerysetMixin(object):
    search_fields = None
    lookup_fields = None
    search_form_class = SearchForm
    filters = []
    method = 'GET'

    def dispatch(self, request, *args, **kwargs):
        self.params = self.get_params(request)
        self.querydict = self.get_querydict(request)
        return super(FilterQuerysetMixin, self).dispatch(request, *args, **kwargs)

    def get_params(self, request):
        querydict = self.get_querydict(request)
        return dict(querydict.items())

    def get_querydict(self, request):
        return getattr(request, self.method)

    @property
    def query_param(self):
        search_form = type(self.get_search_form())
        return filter_dictionary(search_form.base_fields, lambda x: isinstance(x, QueryField))

    @property
    def query(self):
        return self.params.get(self.query_param, '')

    def get_search_form_class(self):
        return self.search_form_class

    def get_search_form(self, *args, **kwargs):
        return self.get_search_form_class()(*args, **kwargs)

    def get_query_help_message(self):
        if not self.search_fields:
            return ''
        fields = []
        for search_field in self.search_fields:
            field = get_field_from_path(self.model, search_field)
            if field is None:
                raise ImproperlyConfigured("cannot access to `%s` from %s " % (search_field, self.model))
            fields.append(field.verbose_name.title())
        return ",".join(fields)

    def get_lookup_fields(self):
        if self.lookup_fields == ANY_LOOKUP_ALLOWED:
            return ANY_LOOKUP_ALLOWED
        lk = self.lookup_fields or []
        form = self.get_search_form()
        if form:
            for attr_name in form.base_fields.keys():
                if not attr_name == self.query_param and not attr_name.startswith('_'):
                    lk.append(attr_name)
        return StartWithList(lk)

    def _get_filters(self):
        filters = {}
        lookup_fields = self.get_lookup_fields()
        if lookup_fields:
            for k, v in self.params.items():
                if v and ((lookup_fields == ANY_LOOKUP_ALLOWED) or (k in lookup_fields)):
                    real_value = self.querydict.getlist(k)
                    if len(real_value) == 1:
                        filters[k] = real_value[0]
                    else:
                        filters['%s__in' % k] = real_value
        return filters

    def _filter_qs(self, qs):
        def construct_search(field_name):
            if field_name.startswith('^'):
                return "%s__istartswith" % field_name[1:]
            elif field_name.startswith('='):
                return "%s__iexact" % field_name[1:]
            elif field_name.startswith('@'):
                return "%s__search" % field_name[1:]
            else:
                return "%s__icontains" % field_name

        filters = self._get_filters()
        if self.search_fields and self.query:
            orm_lookups = [construct_search(str(search_field).replace('.', '__'))
                           for search_field in self.search_fields]
            for bit in self.query.split():
                or_queries = [models.Q(**{orm_lookup: bit})
                              for orm_lookup in orm_lookups]
                qs = qs.filter(reduce(operator.or_, or_queries))

        return qs.filter(**filters).select_related()

    def get_queryset(self):
        qs = super(FilterQuerysetMixin, self).get_queryset()
        return self._filter_qs(qs)

    def get_context_data(self, **kwargs):
        kwargs.update({
            'query': self.query,
            'query_param': self.query_param,
            'searchform': self.get_search_form(self.querydict, initial=dict(self.querydict.items())),
            'query_help_message': self.get_query_help_message(),
            'search_fields': self.search_fields,
        })
        return kwargs



# -*- coding: utf-8 -*-
from collections import OrderedDict
import copy
import logging
from django.core.exceptions import ImproperlyConfigured
from django.utils.functional import curry
from ereports.engine.cache import DatasourceCacheManager, DummyCacheManager
from ereports.engine.columns import Column, get_column_for_attribute, RowValueError
from ereports.engine.utils import get_tables_for_query
from ereports.utils import get_model_field_names

REPR_OUTPUT_SIZE = 20

logger = logging.getLogger(__name__)


class RecordFilteredError(Exception):
    pass


class DatasourceRow(OrderedDict):
    def __getattr__(self, item):
        if item in self:
            return self[item]
        return OrderedDict.__getattr__(self, item)

    def __eq__(self, other):
        return tuple(v.value for k, v in self.items()) == other


class Datasource(object):
    """
        Represent a set of data. Basically act as a queryset,
        but can work with any model attribute and can traverse them.

        Each row is a classnametuple

    :param source can be Model, Manager or Queryset
    """
    model = None
    queryset = None
    columns = None
    extras = None
    order_by = None
    use_cache = False
    dependent_models = None  # used by the cache system
    cache_manager = None

    def __init__(self, **kwargs):
        self.kwfilters = {}
        self.filters = []
        self._result_cache = None
        self._queryset = None
        self._custom_filters = kwargs.pop('custom_filters', [])
        self.extras = kwargs.pop('extras', {})

        for key, value in kwargs.iteritems():
            setattr(self, key, value)

        if self.use_cache and self.cache_manager is None:
            self.cache_manager = DatasourceCacheManager()
        elif not self.use_cache:
            self.cache_manager = DummyCacheManager()

    @classmethod
    def as_datasource(cls, **initkwargs):
        for key, value in initkwargs.items():
            if not hasattr(cls, key) and not callable(value):
                raise TypeError(u"%s() received an invalid keyword %r" % (cls.__name__, key))

        queryset = initkwargs.get('queryset', cls.queryset)
        model = initkwargs.get('model', cls.model)
        _columns = initkwargs.get('columns', cls.columns)

        if queryset:
            model = queryset.model
        elif model is None:
            raise ImproperlyConfigured(u"%(cls)s is missing a queryset. Define "
                                       u"%(cls)s.model, %(cls)s.queryset" % {'cls': cls.__name__})

        factory = curry(get_column_for_attribute, model)
        if not _columns:
            initkwargs['columns'] = [factory(name)(name, model=model)
                                     for name in get_model_field_names(model)]
        else:
            columns = list(_columns)
            for idx, colname in enumerate(_columns):
                if isinstance(colname, basestring):
                    columns[idx] = factory(colname)(colname, model=model)
                elif isinstance(colname, Column):
                    colname.model = model
                elif isinstance(colname, (list, tuple)):
                    name, Class = colname
                    columns[idx] = Class(name, model=model)
            initkwargs['columns'] = columns

        initkwargs['RowClass'] = DatasourceRow
        initkwargs['model'] = model

        return cls(**initkwargs)

    def __repr__(self):
        data = list(self.get_data()[:REPR_OUTPUT_SIZE + 1])
        if len(data) > REPR_OUTPUT_SIZE:
            data[-1] = "...(remaining elements truncated)..."
        return repr(data)

    def __len__(self):
        return len(self.get_data())

    def filter_record(self, obj):
        pass

    def add_custom_filter(self, func):
        self._custom_filters.append(func)

    def __iter__(self):
        return iter(self.get_data())

    def __getitem__(self, k):
        ds = self.get_data()
        ret = ds[k]
        return ret

    def add_filters(self, *args, **kwargs):
        for k, v in kwargs.items():
            self.kwfilters[k] = v
        for v in args:
            self.filters.append(v)

    def _create_result_cache(self):
        result_cache = []
        qs = self._get_queryset()
        for obj in qs:
            try:
                self.filter_record(obj)
                row = self._get_values_from_object(obj)
                for func in self._custom_filters:
                    func(row)
                result_cache.append(row)
            except RecordFilteredError:
                pass
            except ValueError as e:
                logger.exception(e)
        return tuple(result_cache)

    def get_data(self):
        if self._result_cache is None:
            cache_key = self.cache_manager.get_key(self)
            cached_data = self.cache_manager.retrieve(cache_key)
            if cached_data is None:
                self._result_cache = self._create_result_cache()
                self.cache_manager.store(cache_key, self._result_cache)
            else:
                self._result_cache = cached_data
        return self._result_cache

    def _get_queryset(self):
        if not bool(self._queryset):
            if self.queryset:
                qs = self.queryset
            else:
                qs = self.model._default_manager.all()
            qs = qs.filter(*self.filters, **self.kwfilters)
            if self.order_by is not None:
                qs = qs.order_by(*self.order_by)
            qs = qs.select_related()
            self._queryset = qs
            self.dependent_tables = get_tables_for_query(qs.query)
        return self._queryset

    def query(self):
        self.get_data()
        return self._queryset.query

    def _get_values_from_object(self, obj):
        """
            process each queryset entry and returns a list of Column()
        :param obj:
        :return:
        """
        row = self.RowClass()
        for col in self.columns:
            try:
                cell = col.get_value(obj, self)
                row[col.name] = col.apply_manipulator(cell)
                row._original = obj
                # values.append(col.apply_manipulator(cell))
            except Exception as e:
                row[col.name] = RowValueError(e)
                # values.append(RowValueError(e))

        # row = self.RowClass(values)
        return row

    def _clone(self, extras=None):
        klass = self.__class__
        kwargs = copy.deepcopy(self.__dict__)
        if extras is not None:
            kwargs["extras"] = extras
        if self.queryset:
            kwargs.update({'queryset': self.queryset._clone()})
        c = klass()
        c.__dict__.update(kwargs)
        return c



import json
from django import forms
from django.contrib import admin
from django.contrib.auth.models import Permission
from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import ValidationError
from django.core.urlresolvers import reverse
from django.db import models
from django.forms.models import ModelForm
from django.template.context import get_standard_processors
from pasportng.ng.admin import NGModelAdmin
from .models import Report, ReportGroup, ReportTemplate, ReportPermission
from pasportng.pasport.models import PasportOffice
from pasportng.shortcuts.iguardian import GroupObjectPermissionAdmin


class ReportForm(ModelForm):
#    manager = forms.ChoiceField()
    manager = forms.CharField(initial='filter')

    def __init__(self, *args, **kwargs):
        super(ReportForm, self).__init__(*args, **kwargs)
        self.fields['target_model'].queryset = ContentType.objects.order_by('name')
        if 'instance' in kwargs:
            try:
                extras = json.loads(kwargs['instance'].extras)
                self.fields['manager'].initial = extras['manager']
            except ValueError:
                extras = ''

    class Meta:
        model = Report

    def clean(self):
        self.cleaned_data['extras'] = json.dumps({'manager': self.cleaned_data['manager']})
        return super(ReportForm, self).clean()


def publish(modeladmin, request, queryset):
    queryset.update(published=True)


def unpublish(modeladmin, request, queryset):
    queryset.update(published=False)

#class ReportPermission(GroupObjectPermission):
#    pass


class IReport(NGModelAdmin):
    save_on_top = True
    list_display = ('name', 'target_model', 'datasource', 'published', 'template', 'group', 'preview')
    list_filter = cell_filter = ('target_model', 'published', 'template', 'group')
    form = ReportForm
    actions = [publish, unpublish]
    formfield_overrides = {
        models.TextField: {'widget': forms.Textarea({'cols': '20', 'rows': '20'})}, }

    search_fields = ('name', )
    fieldsets = [
        (None, {'fields': (('name', 'target_model', 'published', 'select_related', 'use_distinct'),
                           ('ttl', 'cache_key', 'template', 'group', 'datasource', 'manager'))}),
        ('Columns', {'classes': ['ollapse', 'textarea-fields', ],
                     'fields': [('columns', 'filtering', 'ordering', 'groupby', 'extras'), ]}),
        ('Extra', {'classes': ('collapse textarea-fields',), 'fields': [('rawsql',), ]}),

    ]

    def preview(self, u):
        url = reverse('ereports.do_report', args=[PasportOffice.objects.get(pk=1).code, u.pk, 'html'])
        return '<a target="_new" href="%s">HTML</a>' % url

    preview.allow_tags = True

    def change_view(self, request, object_id, form_url='', extra_context=None):
        object = Report.objects.get(pk=object_id)
        import sqlparse

        ctx = {}
        for cp in get_standard_processors():
            ctx.update(**cp(request))
        ctx.update(**object.get_extra_context())
        try:
            qs = object.get_target_queryset(**ctx)
        except ValidationError:
            qs = object.target_model.model_class().objects.all()

        context = {'qs': qs,
                   'fields': sorted(qs.model._meta.fields, key=lambda x: x.verbose_name),
                   'query': sqlparse.format(str(qs.query), reindent=True, keyword_case='upper')}
        if extra_context:
            context.update(extra_context)

        return super(IReport, self).change_view(request, object_id, form_url, context)


class ReportPermissionForm(ModelForm):
#    content_type = forms.ModelChoiceField( ContentType.objects.all(), initial="something")

    def __init__(self, *args, **kwargs):
        super(ReportPermissionForm, self).__init__(*args, **kwargs)
        self.fields['content_type'].initial = ContentType.objects.get_for_model(Report)
        self.fields['permission'].initial = Permission.objects.get(content_type__app_label='ereports', codename='run_report')

    class Meta:
        model = ReportPermission

class ReportPermissionAdmin(GroupObjectPermissionAdmin):
#    readonly_fields = ['content_type', ]

    form = ReportPermissionForm

    def formfield_for_dbfield(self, db_field, **kwargs):
        request = kwargs.get("request", None)
        formfield = super(ReportPermissionAdmin, self).formfield_for_dbfield(db_field, **kwargs)
        ct = ContentType.objects.get_for_model(Report)
        if db_field.name == 'permission':
            formfield.queryset = Permission.objects.filter(content_type=ct)
        elif db_field.name == 'object_pk':
            formfield = forms.ChoiceField(Report.objects.values_list('id', 'name'))
        elif db_field.name == 'content_type':
            formfield.queryset = ContentType.objects.filter(pk=ct.pk)

        return formfield

    def queryset(self, request):
        ct = ContentType.objects.get_for_model(Report)
        return super(ReportPermissionAdmin, self).queryset(request).filter(content_type=ct)


admin.site.register(Report, IReport)
admin.site.register(ReportPermission, ReportPermissionAdmin)
admin.site.register(ReportTemplate, NGModelAdmin)
admin.site.register(ReportGroup, NGModelAdmin)

__iadmin__ = ((Report, IReport), (ReportTemplate, NGModelAdmin), (ReportGroup, NGModelAdmin), (ReportPermission, ReportPermissionAdmin))
